#!/usr/bin/env python
import argparse
import sys
import os.path
import re
from agouti_pkg.pyfaidx import Fasta, wrap_sequence, FetchError, ucsc_split, bed_split
from collections import defaultdict

keepcharacters = (' ', '.', '_')

def write_sequence(args):
    _, ext = os.path.splitext(args.fasta)
    if ext:
        ext = ext[1:]  # remove the dot from extension

    filt_function = re.compile(args.regex).search

    if args.invert_match:
        filt_function = lambda x: not re.compile(args.regex).search(x)

    fasta = Fasta(args.fasta, default_seq=args.default_seq, key_function=eval(args.header_function), strict_bounds=not args.lazy, split_char=args.delimiter, filt_function=filt_function, read_long_names=args.long_names, rebuild=not args.no_rebuild)

    regions_to_fetch, split_function = split_regions(args)
    if not regions_to_fetch:
        regions_to_fetch = fasta.keys()

    header = False
    for region in regions_to_fetch:
        name, start, end = split_function(region)
        if args.size_range:
            if start is not None and end is not None:
                sequence_len = end - start
            else:
                sequence_len = len(fasta[name])
            if args.size_range[0] > sequence_len or args.size_range[1] < sequence_len:
                continue
        if args.split_files:  # open output file based on sequence name
            filename = '.'.join(str(e) for e in (name, start, end, ext) if e)
            filename = ''.join(c for c in filename if c.isalnum() or c in keepcharacters)
            outfile = open(filename, 'w')
        elif args.out:
            outfile = args.out
        else:
            outfile = sys.stdout
        try:
            if args.transform:
                if not header and args.transform == 'nucleotide':
                    outfile.write("name\tstart\tend\tA\tT\tC\tG\tN\tothers\n")
                    header = True
                outfile.write(transform_sequence(args, fasta, name, start, end))
            else:
                for line in fetch_sequence(args, fasta, name, start, end):
                    outfile.write(line)
        except FetchError as e:
            raise FetchError(str(e) + " Try setting --lazy.\n")
        if args.split_files:
            outfile.close()
    fasta.__exit__()


def fetch_sequence(args, fasta, name, start=None, end=None):
    try:
        line_len = fasta.faidx.index[name].lenc
        if args.auto_strand and start > end and start is not None and end is not None:
            # flip (0, 1] coordinates
            sequence = fasta[name][end - 1:start + 1]
            sequence = sequence.reverse.complement
        else:
            sequence = fasta[name][start:end]
    except KeyError:
        sys.stderr.write("warning: {name} not found in file\n".format(**locals()))
        return
    if args.complement:
        sequence = sequence.complement
    if args.reverse:
        sequence = sequence.reverse
    if args.no_output:
        return
    if args.no_names:
        pass
    else:
        if (start or end) and not args.no_coords:
            yield ''.join(['>', sequence.fancy_name, '\n'])
        else:
            yield ''.join(['>', sequence.name, '\n'])
    for line in wrap_sequence(line_len, sequence.seq):
        yield line


def mask_sequence(args):
    fasta = Fasta(args.fasta, mutable=True, split_char=args.delimiter)

    regions_to_fetch, split_function = split_regions(args)

    for region in regions_to_fetch:
        rname, start, end = split_function(region)
        if args.mask_with_default_seq:
            if start and end:
                span = end - start
            elif not start and not end:
                span = len(fasta[rname])
            else:
                span = len(fasta[rname][start:end])
            fasta[rname][start:end] = span * args.default_seq
        elif args.mask_by_case:
            fasta[rname][start:end] = fasta[rname][start:end].lowercase()


def split_regions(args):
    if args.bed:
        regions_to_fetch = args.bed
        split_function = bed_split
    else:
        regions_to_fetch = args.regions
        split_function = ucsc_split
    return (regions_to_fetch, split_function)


def transform_sequence(args, fasta, name, start=None, end=None):
    line_len = fasta.faidx.index[name].lenc
    s = fasta[name][start:end]
    if args.no_output:
        return
    if args.transform == 'bed':
        return '{name}\t{start}\t{end}\n'.format(name=s.name, start=s.start - 1 , end=s.end)
    elif args.transform == 'chromsizes':
        return '{name}\t{length}\n'.format(name=s.name, length=len(s))
    elif args.transform == 'nucleotide':
        ss = str(s).upper()
        nucs = defaultdict(int)
        nucs.update([(c, str(ss).count(c)) for c in set(str(ss))])
        A = nucs.pop('A', 0)
        T = nucs.pop('T', 0)
        C = nucs.pop('C', 0)
        G = nucs.pop('G', 0)
        N = nucs.pop('N', 0)
        others = '|'.join([':'.join((k, v)) for k, v in nucs.items()])
        return '{sname}\t{sstart}\t{send}\t{A}\t{T}\t{C}\t{G}\t{N}\t{others}\n'.format(sname=s.name, sstart=s.start, send=s.end, **locals())
    elif args.transform == 'transposed':
        return '{name}\t{start}\t{end}\t{seq}\n'.format(name=s.name, start=s.start, end=s.end, seq=str(s))



def main(ext_args=None):
    from pyfaidx import __version__
    parser = argparse.ArgumentParser(description="Fetch sequences from FASTA. If no regions are specified, all entries in the input file are returned. Input FASTA file must be consistently line-wrapped, and line wrapping of output is based on input line lengths.",
                                     epilog="Please cite: Shirley MD, Ma Z, Pedersen BS, Wheelan SJ. (2015) Efficient \"pythonic\" access to FASTA files using pyfaidx. PeerJ PrePrints 3:e1196 https://dx.doi.org/10.7287/peerj.preprints.970v1")
    parser.add_argument('fasta', type=str, help='FASTA file')
    parser.add_argument('regions', type=str, nargs='*', help="space separated regions of sequence to fetch e.g. chr1:1-1000")
    _input = parser.add_argument_group('input options')
    output = parser.add_argument_group('output options')
    header = parser.add_argument_group('header options')
    _input.add_argument('-b', '--bed', type=argparse.FileType('r'), help="bed file of regions")
    output.add_argument('-o', '--out', type=argparse.FileType('w'), help="output file name (default: stdout)")
    output.add_argument('-i', '--transform', type=str, choices=('bed', 'chromsizes', 'nucleotide', 'transposed'), help="transform the requested regions into another format. default: %(default)s")
    output.add_argument('-c', '--complement', action="store_true", default=False, help="complement the sequence. default: %(default)s")
    output.add_argument('-r', '--reverse', action="store_true", default=False, help="reverse the sequence. default: %(default)s")
    output.add_argument('-y', '--auto-strand', action="store_true", default=False, help="reverse complement the sequence when start > end coordinate. default: %(default)s")
    output.add_argument('-a', '--size-range', type=parse_size_range, default=None, help='selected sequences are in the size range [low, high]. example: 1,1000 default: %(default)s')
    names = header.add_mutually_exclusive_group()
    names.add_argument('-n', '--no-names', action="store_true", default=False, help="omit sequence names from output. default: %(default)s")
    names.add_argument('-f', '--long-names', action="store_true", default=False, help="output full (long) names from the input fasta headers. default: headers are truncated after the first whitespace")
    header.add_argument('-t', '--no-coords', action="store_true", default=False, help="omit coordinates (e.g. chr:start-end) from output headers. default: %(default)s")
    output.add_argument('-x', '--split-files', action="store_true", default=False, help="write each region to a separate file (names are derived from regions)")
    output.add_argument('-l', '--lazy', action="store_true", default=False, help="fill in --default-seq for missing ranges. default: %(default)s")
    output.add_argument('-s', '--default-seq', type=check_seq_length, default=None, help='default base for missing positions and masking. default: %(default)s')
    header.add_argument('-d', '--delimiter', type=str, default=None, help='delimiter for splitting names to multiple values (duplicate names will be discarded). default: %(default)s')
    header.add_argument('-e', '--header-function', type=str, default='lambda x: x.split()[0]', help='python function to modify header lines e.g: "lambda x: x.split("|")[0]". default: %(default)s')
    header.add_argument('-u', '--duplicates-action', type=str, default="stop", choices=("stop", "first", "last", "longest", "shortest"), help='entry to take when duplicate sequence names are encountered. default: %(default)s')
    matcher = parser.add_argument_group('matching arguments')
    matcher.add_argument('-g', '--regex', type=str, default='.*', help='selected sequences are those matching regular expression. default: %(default)s')
    matcher.add_argument('-v', '--invert-match', action="store_true", default=False, help="selected sequences are those not matching 'regions' argument. default: %(default)s")
    masking = output.add_mutually_exclusive_group()
    masking.add_argument('-m', '--mask-with-default-seq', action="store_true", default=False, help="mask the FASTA file using --default-seq default: %(default)s")
    masking.add_argument('-M', '--mask-by-case', action="store_true", default=False, help="mask the FASTA file by changing to lowercase. default: %(default)s")
    output.add_argument('--no-output', action="store_true", default=False, help="do not output any sequence. default: %(default)s")
    parser.add_argument('--no-rebuild', action="store_true", default=False, help="do not rebuild the .fai index even if it is out of date. default: %(default)s")
    parser.add_argument('--version', action="version", version=__version__, help="print pyfaidx version number")
    # print help usage if no arguments are supplied
    if len(sys.argv)==1 and not ext_args:
        parser.print_help()
        sys.exit(1)
    elif ext_args:
        args = parser.parse_args(ext_args)
    else:
        args = parser.parse_args()
        
    if args.auto_strand:
        if args.complement:
            sys.stderr.write("--auto-strand and --complement are both set. Are you sure this is what you want?\n")
        if args.reverse:
            sys.stderr.write("--auto-strand and --reverse are both set. Are you sure this is what you want?\n")

    if args.mask_with_default_seq or args.mask_by_case:
        mask_sequence(args)
    else:
        write_sequence(args)

    

def check_seq_length(value):
    if value is None:
        pass # default value
    elif len(value) != 1:
        # user is passing a single character
        raise argparse.ArgumentTypeError("--default-seq value must be a single character!")
    return value

def parse_size_range(value):
    """ Size range argument should be in the form start,end and is end-inclusive. """
    if value is None:
        return value
    try:
        start, end = value.replace(' ', '').replace('\t', '').split(',')
    except (TypeError, ValueError, IndexError):
        raise ValueError
    return (int(start), int(end))


class Counter(dict):
    '''Dict subclass for counting hashable objects.  Sometimes called a bag
    or multiset.  Elements are stored as dictionary keys and their counts
    are stored as dictionary values.
    '''

    def __init__(self, iterable=None, **kwds):
        '''Create a new, empty Counter object.  And if given, count elements
        from an input iterable.  Or, initialize the count from another mapping
        of elements to their counts.
        '''
        self.update(iterable, **kwds)

    def __missing__(self, key):
        return 0

    def most_common(self, n=None):
        '''List the n most common elements and their counts from the most
        common to the least.  If n is None, then list all element counts.
        '''
        if n is None:
            return sorted(self.iteritems(), key=itemgetter(1), reverse=True)
        return nlargest(n, self.iteritems(), key=itemgetter(1))

    def elements(self):
        '''Iterator over elements repeating each as many times as its count.

        If an element's count has been set to zero or is a negative number,
        elements() will ignore it.

        '''
        for elem, count in self.iteritems():
            for _ in repeat(None, count):
                yield elem

    # Override dict methods where the meaning changes for Counter objects.

    @classmethod
    def fromkeys(cls, iterable, v=None):
        raise NotImplementedError(
            'Counter.fromkeys() is undefined.  Use Counter(iterable) instead.')

    def update(self, iterable=None, **kwds):
        '''Like dict.update() but add counts instead of replacing them.

        Source can be an iterable, a dictionary, or another Counter instance.

        '''
        if iterable is not None:
            if hasattr(iterable, 'iteritems'):
                if self:
                    self_get = self.get
                    for elem, count in iterable.iteritems():
                        self[elem] = self_get(elem, 0) + count
                else:
                    dict.update(self, iterable) # fast path when counter is empty
            else:
                self_get = self.get
                for elem in iterable:
                    self[elem] = self_get(elem, 0) + 1
        if kwds:
            self.update(kwds)

    def copy(self):
        'Like dict.copy() but returns a Counter instance instead of a dict.'
        return Counter(self)

    def __delitem__(self, elem):
        'Like dict.__delitem__() but does not raise KeyError for missing values.'
        if elem in self:
            dict.__delitem__(self, elem)

    def __repr__(self):
        if not self:
            return '%s()' % self.__class__.__name__
        items = ', '.join(map('%r: %r'.__mod__, self.most_common()))
        return '%s({%s})' % (self.__class__.__name__, items)

    # Multiset-style mathematical operations discussed in:
    #       Knuth TAOCP Volume II section 4.6.3 exercise 19
    #       and at http://en.wikipedia.org/wiki/Multiset
    #
    # Outputs guaranteed to only include positive counts.
    #
    # To strip negative and zero counts, add-in an empty counter:
    #       c += Counter()

    def __add__(self, other):
        '''Add counts from two counters.

        '''
        if not isinstance(other, Counter):
            return NotImplemented
        result = Counter()
        for elem in set(self) | set(other):
            newcount = self[elem] + other[elem]
            if newcount > 0:
                result[elem] = newcount
        return result

    def __sub__(self, other):
        ''' Subtract count, but keep only results with positive counts.

        '''
        if not isinstance(other, Counter):
            return NotImplemented
        result = Counter()
        for elem in set(self) | set(other):
            newcount = self[elem] - other[elem]
            if newcount > 0:
                result[elem] = newcount
        return result

    def __or__(self, other):
        '''Union is the maximum of value in either of the input counters.

        '''
        if not isinstance(other, Counter):
            return NotImplemented
        _max = max
        result = Counter()
        for elem in set(self) | set(other):
            newcount = _max(self[elem], other[elem])
            if newcount > 0:
                result[elem] = newcount
        return result

    def __and__(self, other):
        ''' Intersection is the minimum of corresponding counts.

        '''
        if not isinstance(other, Counter):
            return NotImplemented
        _min = min
        result = Counter()
        if len(self) < len(other):
            self, other = other, self
        for elem in filter(self.__contains__, other):
            newcount = _min(self[elem], other[elem])
            if newcount > 0:
                result[elem] = newcount
        return result

if __name__ == "__main__":
    main()



from operator import attrgetter
from agouti_pkg.sequence_ontology import *
from io import StringIO
import pandas as pd
from agouti_pkg.eprint import eprint


def statistics(agouti_output : str):
    """Create statistics based on the agouti output

    Args:
        agouti_output (str): output of the agouti software
    """
    
    output = StringIO(agouti_output)
    df = pd.read_csv(output, sep="\t", index_col=False) 
    
    columns = df.columns
    
    #check the index of the first column for which stats will be calculated
    for i in range(0, len(columns)):
        if columns[i].startswith("annotated_") and not columns[i+1].startswith("annotated_"):
            n = i+1
            break
    cols = list(columns[n:])
    if "annotated_featuretype" in columns:
        cols.append("annotated_featuretype")
    eprint("##"*10)
    eprint("##STATISTICS")
    eprint("##"*10)
    for col in cols:
        try:
            df[col] = df[col].str.strip()
        except AttributeError:
            pass
        output = []
        col_stats = df[col].value_counts(dropna=False).to_dict()
        for key, value in col_stats.items():
            output.append(f"'{key}' - {value}")
        eprint(f"# statistics for the '{col}' column: {'; '.join(output)}")


def sum_features_length(features):
    """Returns the sum of the features lengths. Features should be of Feature\
        type

    Arguments:
        features {list} -- list or other iterable object of Features

    Returns:
        int -- sum of features lengths
    """

    length = 0
    for f in features:
        length += (f.end - f.start)
    return length


def get_level1_parent(db, child, id=False):
    """Returns 1st level parent or parent id of a given Feature.
        If child has no parrent child will be returned

    Arguments:
        db {modules.database.Database} -- Database object
        child {Feature} -- Feature object

    Keyword Arguments:
        id {bool} -- if True returns id of the object instead of Feature\
                     object (default: {False})

    Returns:
        Feature -- child Feature or id if id parameter is set to True
    """

    for p in db.database.parents(child):
        p_parents = db.database.parents(p)
        if sum(1 for a in db.database.parents(p)) == 0:
            return p.id if id else p
        elif sum(1 for c in db.database.parents(p)) == 1:
            for a in db.database.parents(p):
                if p.featuretype == a.featuretype:
                    return p.id if id else p
    return child.id if id else child


def completly_within(start, end, feature):
    """Checks if product described by start and end coordinates is completly\
       within feature.

    Arguments:
        start {int} -- start coordinate of annotated product
        end {int} -- end coordinate of annotated product
        feature {Feature} -- Feature object

    Returns:
        bool -- True if product is completly within Feature object,\
                False otherwise
    """

    return True if start >= feature.start and end <= feature.end else False


def get_min_start(featurelist):
    """Returns the smallest feature start coordinate from the list of Feature\
        objects

    Arguments:
        featurelist {list} -- list of Feature objects

    Returns:
        int -- min start coordinate from a given set
    """

    starts = []
    for feature in featurelist:
        starts.append(feature.start)
    return min(starts)


def get_max_end(featurelist):
    """Returns the biggest feature end coordinate from the list of Feature\
        objects

    Arguments:
        featurelist {list} -- list of Feature objects

    Returns:
        int -- max end coordinate from a given set
    """

    ends = []
    for feature in featurelist:
        ends.append(feature.start)
    return max(ends)


def distinguish_UTRs(utr, cds):
    """Distinguish 3' from 5' UTRs if not annotated

    Arguments:
        utr {list} -- list of UTRs
        cds {list} -- list of CDS

    Returns:
        tuple -- tuple of sets containing 3' and 5' UTRs
    """

    cds_start, cds_end = get_min_start(cds), get_max_end(cds)
    three_primes, five_primes = set(), set()
    for u in utr:
        if (((u.end <= cds_start) and u.strand == "+")
           or ((u.start >= cds_end) and u.strand == "-")):
            five_primes.add(u)
        elif (((u.start >= cds_end) and u.strand == "+")
              or ((u.end <= cds_start) and u.strand == "-")):
            three_primes.add(u)
    return (five_primes, three_primes)


def get_exons(feature, db):
    """Returns exons of a given transcript.

    Arguments:
        feature {Feature} -- transcript
        db {modules.database.Database} -- Database object

    Returns:
        list -- list of exons
    """

    utr5_children = db.get_children(feature, featuretypes=["exon"], level=1)
    exons_children = db.get_children(feature, featuretypes=["exon"], level=1)
    return list(exons_children)


def get_cds(feature, featuretypes_from_db, db):
    """Returns CDS of a given transcript

    Arguments:
        feature {Feature} -- transcript
        featuretypes_from_db {list} -- list of feature types from the database
        db {modules.database.Database} -- Database object

    Returns:
        tuple -- tuple of CDS
    """

    synonyms = cds_synonyms
    intersection = set(synonyms) & set(featuretypes_from_db)
    children = db.get_children(feature, featuretypes=intersection, level=1)
    return children


def get_three_prime_UTRs(feature, featuretypes_from_db, db):
    """Returns 3' UTR regions of a given transcript

    Arguments:
        feature {Feature} -- transcript
        featuretypes_from_db {list} -- list of feature types from the database
        db {modules.database.Database} -- Database object

    Returns:
        list -- list of 3 prime UTRs. Usually of length equal to 1
    """

    synonyms = [three_prime_UTR_synonyms, UTR_synonyms]
    intersection = set(synonyms[0]) & set(featuretypes_from_db)
    utr3_children = db.get_children(feature, featuretypes=intersection,
                                    level=1)
    intersection = set(synonyms[1]) & set(featuretypes_from_db)
    utr_children = db.get_children(feature, featuretypes=intersection, level=1)
    cds = get_cds(feature, featuretypes_from_db, db)
    five_primes, three_primes = [], []
    if len(utr_children) and len(cds) and len(intersection):
        five_primes, three_primes = distinguish_UTRs(utr_children, cds)

    result_list = []
    try:
        for a in (list(utr3_children) + list(three_primes)):
            if "+" in feature.strand:
                if not a.start < get_max_end(cds):
                    result_list.append(a)
            elif "-" in feature.strand:
                if not a.start > get_min_start(cds):
                    result_list.append(a)
    except ValueError:
        return []
    return result_list


def get_five_prime_UTRs(feature, featuretypes_from_db, db):
    """Returns 5' UTR regions of a given transcript

    Arguments:
        feature {Feature} -- transcript
        featuretypes_from_db {list} -- list of feature types from the database
        db {modules.database.Database} -- Database object

    Returns:
        list -- list of 3 prime UTRs. Usually of length equal to 1
    """

    synonyms = [five_prime_UTR_synonyms, UTR_synonyms]
    intersection = set(synonyms[0]) & set(featuretypes_from_db)
    utr5_children = db.get_children(
        feature, featuretypes=intersection, level=1)
    intersection = set(synonyms[1]) & set(featuretypes_from_db)
    utr_children = db.get_children(feature, featuretypes=intersection, level=1)
    cds = get_cds(feature, featuretypes_from_db, db)

    five_primes, three_primes = [], []
    if len(utr_children) and len(cds) and len(intersection):
        five_primes, three_primes = distinguish_UTRs(utr_children, cds)

    result_list = []
    try:
        for a in (list(utr5_children) + list(five_primes)):
            if "+" in feature.strand:
                if not a.start > get_min_start(cds):
                    result_list.append(a)
            elif "-" in feature.strand:
                if not a.start < get_max_end(cds):
                    result_list.append(a)
    except ValueError:
        return []
    return result_list


def handle_negative_coordinates(start, end):
    """Checks if coordinates are outside transcript boundaries

    Arguments:
        start {int} -- start coordinate
        end {int} -- end coordinate

    Returns:
        tuple -- updated start coordinate, updated end coordinate,
                 string describing which coordinates are outside
                 transcript boundaries
    """

    coord_outside_transcript = "No"

    if (start < 0 and end > 0):
        start_coord, end_coord = 0, end
        coord_outside_transcript = "just_one"
    elif (start < 0 and end < 0):
        start_coord, end_coord = 0, 0
        coord_outside_transcript = "both"
    else:
        start_coord, end_coord = start, end

    return start_coord, end_coord, coord_outside_transcript


def infer_feature_level(feature, db):
    """Infers feature level

    Arguments:
        feature {Feature} -- feature which level is to be inferred
        db {Database} -- object of the Database class

    Returns:
        int -- level
    """

    level = 1
    for parent in db.database.parents(feature):
        if (parent.featuretype != feature.featuretype):
            level = level + 1
    return level


def completly_within_positive_coordinates(lengths_dict, feature,
                                          processing_product, cds_start):
    """Checks if coordinates of ProcessingProduct object don't exceed\
        transcript length.

    Arguments:
        lengths_dict {dict} -- dictionary of utr and cds lengths
        feature {Feature} -- feature, transcript
        processing_product {ProcessingProduct} -- object of the\
            ProcessingProduct class
        cds_start {int} -- start of the CDS

    Returns:
        tuple -- ...
    """

    length = sum(lengths_dict[feature.id])
    five_prime_UTR_len = lengths_dict[feature.id][0]
    if (processing_product.coordinates[2] >
            length and processing_product.coordinates[1] > length):
        return ("both")
    elif (processing_product.coordinates[2] > length):
        return ("just_one", length, five_prime_UTR_len)
    return ("No", length, five_prime_UTR_len)


def add_quotation_marks(l):
    """Add quotation marks for each string in list

    Arguments:
        l {list} -- list of strings

    Returns:
        [list] -- returns list of strings with added quotation marks
    """

    return ["'" + x + "'" for x in l]


def check_position_within_transcript(args, lengths, num, processing_product):
    """Checks whether processing_product lies within 3' UTR, CDS or 5' UTR

    Arguments:
        args {argparse.Namespace} -- argparse command-line arguments
        lengths {tuple} -- 5'UTR length, CDS length, 3'UTR length
        num {int} -- index of the feature. Might be one 0, 1 or 2
        processing_product {ProcessingProduct} -- object of ProcessingProduct\
            class

    Returns:
        bool -- True if ProcessingProduct lies within 3'UTR, 5'UTR or CDS
    """
    
    if (args.completly_within):

        if (num == 0 and processing_product.coordinates[1] >= args.offset and
                processing_product.coordinates[2] < args.offset + lengths[0]):
            # if inside 5'UTR
            return True
        # if inside CDS
        elif (num == 1 and processing_product.coordinates[1] >= args.offset +
              lengths[0] and processing_product.coordinates[2] <=
              (args.offset + lengths[0] + lengths[1])):
            return True
        # if inside 3'UTR
        elif (num == 2 and processing_product.coordinates[1] > (args.offset +
              lengths[0] + lengths[1]) and processing_product.coordinates[2] <
              args.offset + lengths[0] + lengths[1] + lengths[2]):
            return True
        else:
            return False

    else:

        if (num == 0 and processing_product.coordinates[2] > args.offset
            and processing_product.coordinates[1] < args.offset + lengths[0]):
            return True
        elif (num == 1 and (processing_product.coordinates[1] in
              range(args.offset + lengths[0] + 1, args.offset + lengths[0] +
              lengths[1])   
              or processing_product.coordinates[2] in
              range(args.offset + lengths[0] + 1, args.offset +
              lengths[0] + lengths[1])
              or processing_product.coordinates[1] in
              range(args.offset + lengths[0] + 1, args.offset +
              lengths[0] + lengths[1])
              or (processing_product.coordinates[1] <=
              args.offset + lengths[0] and processing_product.coordinates[2] >=
              args.offset + lengths[0] + lengths[1]))):
            return True
        elif (num == 2 and processing_product.coordinates[2] > args.offset +
              lengths[0] + lengths[1] and processing_product.coordinates[1] <
              args.offset + lengths[0] + lengths[1] + lengths[2]):
            return True
        else:
            return False


def find_closest_gene(database, processing_product):
    """Find the closest gene to the processing_product

    Arguments:
        database {Database} -- Database object
        processing_product {ProcessingProduct} -- ProcessingProduct object

    Returns:
        [str] -- string to display in the output file
    """

    f1 = ", ".join(add_quotation_marks(list(database.features_at_1_level)))
    if "-" not in processing_product.strand:

        upstream = database.database.execute(
            'SELECT max(end), id from features where seqid == "{}" AND \
            featuretype IN ({}) AND end <= {}'.format(
                processing_product.coordinates[0], f1,
                processing_product.coordinates[1]))

        downstream = database.database.execute(
            'SELECT min(start), id from features where seqid == "{}" AND \
            featuretype IN ({}) AND start >= {}'.format(
                processing_product.coordinates[0], f1,
                processing_product.coordinates[2]))

    else:
        downstream = database.database.execute(
            'SELECT max(end), id from features where seqid == "{}" AND \
            featuretype IN ({}) AND end <= {}'.format(
                processing_product.coordinates[0], f1,
                processing_product.coordinates[1]))

        upstream = database.database.execute(
            'SELECT min(start), id from features where seqid == "{}" AND \
            featuretype IN ({}) AND start >= {}'.format(
                processing_product.coordinates[0], f1,
                processing_product.coordinates[2]))

    values = []

    for i in upstream.fetchone():
        values.append(i)

    for i in downstream.fetchone():
        values.append(i)

    if "-" not in processing_product.strand:

        return "closest gene upstream: {} - distance {} bp; closest gene downstream: {} - distance {} bp".format(values[1], processing_product.coordinates[1] -
                         values[0] if values[0] is not None else None,
                         values[3], values[2] -
                         processing_product.coordinates[2]
                         if values[2] is not None else None)
    else:

        return "closest gene upstream: {} - distance {} bp; closest gene downstream: {} - distance {} bp".format(values[1], values[0]-processing_product.coordinates[2]
                         if values[0] is not None else None, values[3],
                         processing_product.coordinates[1] - values[2]
                         if values[2] is not None else None)


def filter_overlapping_features(attributes_and_features, args, db, feature,
                                overlapping_features, header_features,
                                header_attr, processing_product=None,
                                lengths=None):
    """Filter overlapping features so that they meet criteria specified by\
        user.
       Parse the results to be displayed in the output file.

    Arguments:
        attributes_and_features {dict} -- dictionary of attributes and\
            features that need to be annotated
        args {argparse.Namespace} -- argparse command-line arguments
        db {Database} -- Database object
        feature {Feature} -- single overlapping Feature object
        overlapping_features {list} -- list of overlapping features
        header_features {list} -- features present in the header
        header_attr {list} -- attributes present in the header

    Keyword Arguments:
        processing_product {ProcessingProduct} -- object of the
            ProcessingProduct class (default: {None})
        lengths {dict} -- dictionary of utr and cds lengths (default: {None})

    Returns:
        str -- results to be displayed
    """

    children_ids = db.get_children(feature, ids=True)
    d = {}
    position = check_position_within_transcript  # abbreviation

    for h in (header_features + header_attr):
        d[h] = "."
    

    if (lengths and feature.featuretype in mRNA_synonyms):  # if transcriptomic

        if lengths[1] != 0:

            for f in d.keys():

                if f in three_prime_UTR_synonyms and lengths[2] != 0:

                    d[f] = "y" if position(args, lengths, 2,
                                            processing_product) else "."

                elif f in five_prime_UTR_synonyms and lengths[1] != 0:

                    d[f] = "y" if position(args, lengths, 0,
                                            processing_product) else "."

                elif f in UTR_synonyms:

                    d[f] = "y" if (
                        position(args, lengths, 2, processing_product) and
                        lengths[2] != 0) or (position(args, lengths, 0,
                                                processing_product) and
                                                lengths[0] != 1) else "."

                elif f in cds_synonyms:
                    d[f] = "y" if position(args, lengths, 1,
                                            processing_product) else "."

        else:

            for f in d.keys():

                if (f in three_prime_UTR_synonyms or
                   f in five_prime_UTR_synonyms or f in UTR_synonyms or
                   f in cds_synonyms):
                    d[f] = "NA"

    else:

        for o in overlapping_features:

            if (o.id in children_ids and
               o.featuretype in attributes_and_features.keys()):

                d[o.featuretype] = "y"

    # attributes
    analyzed_set = set(attributes_and_features[feature.featuretype]
                       ).intersection(list(feature.attributes))
    for a in header_attr:

        for attr in analyzed_set:
            if attr == a:

                d[attr] = ("").join(feature[attr])
    # concatenation
    result = ""

    for h in (header_features + header_attr):
        result += "{}\t".format(d[h])

    return result.strip()




def commonancestors(*nodes):
    """
    Determine common ancestors of `nodes`.

    >>> from anytree import Node
    >>> udo = Node("Udo")
    >>> marc = Node("Marc", parent=udo)
    >>> lian = Node("Lian", parent=marc)
    >>> dan = Node("Dan", parent=udo)
    >>> jet = Node("Jet", parent=dan)
    >>> jan = Node("Jan", parent=dan)
    >>> joe = Node("Joe", parent=dan)

    >>> commonancestors(jet, joe)
    (Node('/Udo'), Node('/Udo/Dan'))
    >>> commonancestors(jet, marc)
    (Node('/Udo'),)
    >>> commonancestors(jet)
    (Node('/Udo'), Node('/Udo/Dan'))
    >>> commonancestors()
    ()
    """
    ancestors = [node.ancestors for node in nodes]
    common = []
    for parentnodes in zip(*ancestors):
        parentnode = parentnodes[0]
        if all([parentnode is p for p in parentnodes[1:]]):
            common.append(parentnode)
        else:
            break
    return tuple(common)




SCHEMA = """

CREATE TABLE features (
    id text,
    seqid text,
    source text,
    featuretype text,
    start int,
    end int,
    score text,
    strand text,
    frame text,
    attributes text,
    extra text,
    bin int,
    primary key (id)
    );

CREATE TABLE relations (
    parent text,
    child text,
    level int,
    primary key (parent, child, level)
    );

CREATE TABLE meta (
    dialect text,
    version text
    );

CREATE TABLE directives (
    directive text
    );

CREATE TABLE autoincrements (
    base text,
    n int,
    primary key (base)
    );

CREATE TABLE duplicates (
    idspecid text,
    newid text,
    primary key (newid)
    );


"""

default_pragmas = {
    'synchronous': 'NORMAL',
    'journal_mode': 'MEMORY',
    'main.page_size': 4096,
    'main.cache_size': 10000,
}

_keys = ['id', 'seqid', 'source', 'featuretype', 'start', 'end', 'score',
         'strand', 'frame', 'attributes', 'extra', 'bin']
_gffkeys = ['seqid', 'source', 'featuretype', 'start', 'end', 'score',
            'strand', 'frame', 'attributes']
_gffkeys_extra = _gffkeys + ['extra']

_SELECT = "SELECT " \
    + ', '.join(_keys) \
    + ", features.rowid as file_order FROM features "

_INSERT = "INSERT INTO features (" \
    + ', '.join(_keys) + ") VALUES (" + ','.join(list('?' * len(_keys))) + ")"


_update_clause = ','.join(['%s = ?' % i for i in _keys])
_UPDATE = "UPDATE features SET " + _update_clause + " WHERE id = ?"


# TODO: create indexes once profiling figures out which ones work best....
INDEXES = []


# This dictionary keeps track of idiosyncracies to [attempt to] maintain
# invariance of file->db->file round trips.
dialect = {

    # Initial semicolon, e.g.,
    #
    #   ;ID=001;
    # vs
    #   ID=001;
    'leading semicolon': False,

    # Semicolon after the last value, e.g.,
    #
    #   ID=001; Name=gene1;
    # vs
    #   ID=001; Name=gene1
    'trailing semicolon': False,

    # e.g.,
    #
    #   gene_id "GENE1"
    # vs
    #   gene_id GENE1
    'quoted GFF2 values': False,

    # Sometimes there's extra space surrounding the semicolon, e.g.,
    #
    #   ID=001;Name=gene1
    # vs
    #   ID=001; Name=gene1
    'field separator': ';',

    # Usually "=" for GFF3; " " for GTF, e.g.,
    #
    #   gene_id "GENE1"
    # vs
    #   gene_id="GENE1"
    'keyval separator': '=',

    # Usually a comma, e.g.,
    #
    #   Parent=gene1,gene2,gene3
    'multival separator': ',',

    # General GTF or GFF format
    'fmt': 'gff3',

    # How multiple values for the same key are handled, e.g.,
    #
    #   Parent=gene1; Parent=gene2;
    # vs
    #   Parent=gene1,gene2;
    #
    # (the first one has repeated keys)
    'repeated keys': False,

    # If these keys exist, then print them in this order.
    'order': ['ID', 'Name', 'gene_id', 'transcript_id'],

}

always_return_list = True
ignore_url_escape_characters = False

# these keyword args are used by iterators.
_iterator_kwargs = (
    'data',
    'checklines', 'transform', 'force_dialect_check', 'dialect', 'from_string')



from agouti_pkg.miscallaneous import *


class ProcessingProduct(object):
    """Feature from the BED file. One line in BED corresponds\
        to one ProcessingProduct"""

    def __init__(self, coordinates, processing_product, score, strand,
                 bed_line, first_base_num, coords_outside_transcript="No"):
        self.coordinates = coordinates
        if (first_base_num == 0):
            one_based_coordinates = (coordinates[0], coordinates[1]+1,
                                     coordinates[2])
            self.coordinates = one_based_coordinates

        self.processing_product = processing_product
        self.score = score
        self.strand = strand
        self.bed_line = bed_line
        self.coords_outside_transcript = coords_outside_transcript

    def convert_coordinates(self):
        """Converts coordinates into UCSC-like format

        Returns:
            str -- coordinates in UCSC-like format
        """

        coord_formated = "{chr}:{start}-{end}".format(
            chr=self.coordinates[0],
            start=self.coordinates[1],
            end=self.coordinates[2])

        return coord_formated

    def genomic_to_transcriptomic(self, database, featuretypes_from_db):
        """Converts genomic coordinates to transcriptomic coordinates of given\
            mRNA. Require two args. One of them is a sqlite3 database created \
            by gffutils, second is a generator object that yields featuretypes\
            from database (as strings)

        Arguments:
            database {modules.database.Database} -- Database object
            featuretypes_from_db {list} -- feature types in the database

        Returns:
            tuple -- first element is tuple of (5' utr len, cds len ,\
                     3' utr len), second is start coordinate of CDS within\
                     transcript
        """

        transcript_id = self.coordinates[0]

        mRNA = database.database[transcript_id]

        three_prime_UTRs = get_three_prime_UTRs(mRNA,
                                                list(featuretypes_from_db),
                                                database)

        five_prime_UTRs = get_five_prime_UTRs(mRNA, list(featuretypes_from_db),
                                              database)

        exons = get_exons(mRNA, database)
        cds = get_cds(mRNA, list(featuretypes_from_db), database)
        five_prime_utr_len = sum_features_length(five_prime_UTRs)
        three_prime_utr_len = sum_features_length(three_prime_UTRs)
        transcript_length = sum_features_length(exons)
        cds_len = transcript_length - five_prime_utr_len - three_prime_utr_len

        if len(cds) == 0:
            cds_start = None
        else:
            cds_start = min(cds, key=attrgetter('start')).start

        if (sum([cds_len, three_prime_utr_len, five_prime_utr_len]) !=
           transcript_length and len(cds)):

            left_side_of_cds_len = 0
            for exon in exons:
                if (cds_start >= exon.end):
                    left_side_of_cds_len += exon.end - exon.start
                elif (cds_start < exon.end and cds_start > exon.start):
                    left_side_of_cds_len += cds_start - exon.start

            five_prime_utr_len = left_side_of_cds_len
            three_prime_utr_len = (transcript_length - five_prime_utr_len
                                   - cds_len)

        lengths_tuple = (five_prime_utr_len, cds_len,
                         three_prime_utr_len)

        if lengths_tuple == (0, 0, 0):
            lengths_tuple = (0, transcript_length, 0)

        return lengths_tuple, cds_start

    def check_overlapping_feature__position(self, lengths_dict, feature,
                                            transcriptomic, offset=0):
        """Checks in which part of overlapping feature lies ProcessingProduct\
            object

        Arguments:
            feature {Feature} -- overlapping feature
            transcriptomic {bool} -- transcriptomic option (see help)

        Keyword Arguments:
            offset {int} -- offset option (see help) (default: {0})

        Returns:
            str -- localization within overlapping feature
        """
        
        start, end = min(self.coordinates[1], self.coordinates[2]), max(self.coordinates[1], self.coordinates[2]) 

        if (transcriptomic):
            feature_length = sum(lengths_dict[feature.id])
        else:
            feature_length = max(feature.stop, feature.start) - min(feature.stop, feature.start)

        pp_length = self.coordinates[2] - self.coordinates[1]

        if (offset):
            feature_start = offset
        else:
            feature_start = 0


        if start > feature_length and end > feature_length:
            loc = "downstream"
        elif self.coords_outside_transcript == "both":
            loc = "upstream"
        elif (start < feature_start + (0.25 * feature_length)
                and end < feature_start + (0.5 * feature_length)):
            loc = "5 prime"
        elif (start > feature_start + (0.25 * feature_length) and
              end < feature_start + (0.75 * feature_length)):
            loc = "middle"
        elif (start > feature_start + (0.5 * feature_length) and
              end > feature_start + (0.75 * feature_length)):
            loc = "3 prime"
        elif (start < feature_start + (0.25 * feature_length) and
              end > feature_start + (0.75 * feature_length) and
              pp_length < 0.9 * feature_length):
            loc = "whole"
        elif (start < feature_start + (0.25 * feature_length) and
              end > feature_start + (0.75 * feature_length) and
              pp_length >= 0.9 * feature_length):
            loc = "full"
        else:
            loc = "other"
        return loc



import unittest
from agouti.gffutils import helpers
from agouti.gffutils import feature


class Test(unittest.TestCase):
    
    def test_merge_attributes(self):
        """
        Tests all possible cases of merging two dictionaries together
        """
        x = {'foo': [1], "baz": 1, "buz": [1], "biz": 1, "boo": [1]}
        y = {'bar': [2], "baz": 2, "buz": [2], "biz": 1, "boo": [1]}
        test = helpers.merge_attributes(x, y)

        for k, v in list(test.items()):
            test[k] = sorted(v)

        true = {'foo': [1],
                'bar': [2],
                "baz": [1, 2],
                "boo": [1],
                "buz": [1, 2],
                "biz": [1]}
        self.assertDictEqual(test, true) 

    def test_merge_Attributes(self):
        f1 = feature.feature_from_line('chr2L . testing 1 10 . + . foo=1; baz=1; buz=1; biz=1; boo=1;', strict=False)
        f2 = feature.feature_from_line('chr2L . testing 1 10 . + . bar=2; baz=2; buz=2; biz=1; boo=1;', strict=False)
        test = helpers.merge_attributes(f1.attributes, f2.attributes)

        for k, v in list(test.items()):
            test[k] = sorted(v)

        true = {'foo': ['1'],
                'bar': ['2'],
                "baz": ['1', '2'],
                "boo": ['1'],
                "buz": ['1', '2'],
                "biz": ['1']}
        self.assertDictEqual(test, true)




# -*- coding: utf-8 -*-
from agouti_pkg.anytree import AnyNode


def _get_indentation(line):
    content = line.lstrip(' ')
    # Split string using version without indentation; First item of result is the indentation itself.
    indentation_length = len(line.split(content)[0])
    return indentation_length, content


class IndentedStringImporter(object):

    def __init__(self, nodecls=AnyNode):
        u"""
        Import Tree from a single string (with all the lines) or list of strings (lines) with indentation.

        Every indented line is converted to an instance of `nodecls`. The string (without indentation) found on the lines are set as the respective node name.

        This importer do not constrain indented data to have a definite number of whitespaces (multiple of any number). Nodes are considered child of a parent simply if its indentation is bigger than its parent.

        This means that the tree can have siblings with different indentations, as long as the siblings indentations are bigger than the respective parent (but not necessarily the same considering each other).

        Keyword Args:
            nodecls: class used for nodes.

        Example using a string list:

        >>> from anytree.importer import IndentedStringImporter
        >>> from anytree import RenderTree
        >>> importer = IndentedStringImporter()
        >>> lines = [
        ...    'Node1',
        ...    'Node2',
        ...    '    Node3',
        ...    'Node5',
        ...    '    Node6',
        ...    '        Node7',
        ...    '    Node8',
        ...    '        Node9',
        ...    '      Node10',
        ...    '    Node11',
        ...    '  Node12',
        ...    'Node13',
        ...]
        >>> root = importer.import_(lines)
        >>> print(RenderTree(root))
       AnyNode(name='root')
        ├── AnyNode(name='Node1')
        ├── AnyNode(name='Node2')
        │   └── AnyNode(name='Node3')
        ├── AnyNode(name='Node5')
        │   ├── AnyNode(name='Node6')
        │   │   └── AnyNode(name='Node7')
        │   ├── AnyNode(name='Node8')
        │   │   ├── AnyNode(name='Node9')
        │   │   └── AnyNode(name='Node10')
        │   ├── AnyNode(name='Node11')
        │   └── AnyNode(name='Node12')
        └── AnyNode(name='Node13')

        Example using a string:

        >>> string = "Node1\n  Node2\n  Node3\n    Node4"
        >>> root = importer.import_(string)
        >>> print(RenderTree(root))
         AnyNode(name='root')
        └── AnyNode(name='Node1')
            ├── AnyNode(name='Node2')
            └── AnyNode(name='Node3')
                └── AnyNode(name='Node4')
        """
        self.nodecls = nodecls

    def _tree_from_indented_str(self, data):
        if isinstance(data, str):
            lines = data.splitlines()
        else:
            lines = data
        root = self.nodecls(name="root")
        indentations = {}
        for line in lines:
            current_indentation, name = _get_indentation(line)

            if len(indentations) == 0:
                parent = root
            elif current_indentation not in indentations:
                # parent is the next lower indentation
                keys = [key for key in indentations.keys() if key < current_indentation]
                parent = indentations[max(keys)]
            else:
                # current line uses the parent of the last line with same indentation and replaces
                # it as the last line with this given indentation
                parent = indentations[current_indentation].parent

            indentations[current_indentation] = self.nodecls(name=name, parent=parent)

            # delete all higher indentations
            keys = [key for key in indentations.keys() if key > current_indentation]
            for key in keys:
                indentations.pop(key)
        return root

    def import_(self, data):
        """Import tree from `data`, which can be a single string or a list of lines."""
        return self._tree_from_indented_str(data)



#!/usr/bin/env python
# coding=utf-8

from agouti_pkg.eprint import eprint
import sys
import os
import pickle
import sqlite3
import agouti_pkg.gffutils
import agouti_pkg.gffutils.inspect as inspect
import argparse
from agouti_pkg.anytree import Node, RenderTree, importer
from agouti_pkg.anytree.exporter import DotExporter
from agouti_pkg.argument_parser import parse_arguments
import gzip

output_lines = []


def transform_func(x):
    x.featuretype = x.featuretype.lower()
    return x


def inspect_db(args, db):
    try:
        features = {}
        if args.annotation.endswith(".gz"):
            f = gzip.open(args.annotation, "rb")
        else:
            f = open(args.annotation)
        lines = f.readlines()
        for line in lines:
            if isinstance(line, bytes):
                line = line.decode()
            splitted = line.strip().split("\t")
            if (not line.startswith("#")):
                feature = splitted[2].lower()
                attributes = splitted[8].strip().replace('"', "").replace("'", "")
                if feature not in features.keys():
                    features[feature] = set()
                tab = attributes.split(db.dialect["field separator"])
                tab = list(filter(None, tab))
                for attribute in tab:
                    a = attribute.strip().split(db.dialect["keyval separator"])
                    a = list(filter(None, a))
                    features[feature].add(a[0])  # was a[0].lower()
        f.close()
    except IndexError:
        eprint("{}ERROR: The file {} has the wrong format".format(os.linesep,
                                                          args.annotation))
        os.system("rm -f {}".format(args.database))
        sys.exit()
    return features


def find_children(_parent, parent_child, prefix=""):
    pre = prefix+"    "
    output_lines.append("{}{}".format(prefix, _parent))
    for (parent, child) in parent_child:
        if _parent == parent:
            find_children(child, parent_child, pre)


def find_roots(x):
    """takes parent_child list of tuples as an argument. The list should be in\
        the format [(parent, child), (parent, child)], etc."""
    roots = set()
    parent_child_relation = list(zip(*x))
    for parent in parent_child_relation[0]:
        if (parent not in parent_child_relation[1]):
            roots.add(parent)
    return roots


def add_super_feature(parent_child):
    """Function adds super-feature if there are multiple first level features"""
    unzipped = list(zip(*parent_child))
    modified_parent_child = set()
    first_level_features = []
    for parent in unzipped[0]:
        if parent not in unzipped[1]:
            first_level_features.append(parent)
    if len(first_level_features) > 1:  # if more than one root
        for i in first_level_features:
            modified_parent_child.add((".", i))
    return parent_child.union(modified_parent_child)


def main(args):
    """Main function

    Arguments:
        argv {argparse.Namespace} -- command-line arguments parsed with\
            argparse
    """

    global output_lines

    try:
        if args.format == 'GFF3':
            if (not args.save_on_disk):
                db = agouti_pkg.gffutils.create_db(args.annotation, force=True,
                                        keep_order=False,
                                        sort_attribute_values=False,
                                        merge_strategy="create_unique",
                                        transform=transform_func,
                                        dbfn=":memory:",
                                        checklines=50)
            else:
                db = agouti_pkg.gffutils.create_db(args.annotation, force=True,
                                        keep_order=False,
                                        sort_attribute_values=False,
                                        merge_strategy="create_unique",
                                        transform=transform_func,
                                        dbfn=args.database,
                                        checklines=50)
        else:
            if (not args.save_on_disk):
                db = agouti_pkg.gffutils.create_db(args.annotation, force=True,
                                    keep_order=False,
                                    sort_attribute_values=False,
                                    merge_strategy="create_unique",
                                    disable_infer_genes= not args.infer_genes,
                                    disable_infer_transcripts=not args.infer_transcripts,
                                    transform=transform_func,
                                    dbfn=":memory:", checklines=50)
            else:
                db = agouti_pkg.gffutils.create_db(args.annotation, force=True,
                                    keep_order=False,
                                    sort_attribute_values=False,
                                    merge_strategy="create_unique",
                                    disable_infer_genes= not args.infer_genes,
                                    disable_infer_transcripts=not args.infer_transcripts,
                                    transform=transform_func,
                                    dbfn=args.database, checklines=50)

    except ValueError:
        eprint("{}ERROR: The file {} has the wrong format or does not exist".format(os.linesep, args.annotation))
        sys.exit()

    database_struct = open("{}.database.structure.txt".format(args.database), "w")

    database_struct.write("{}Attributes available for each feature type:{}\n".format(os.linesep, os.linesep))
    attributes_and_features = inspect_db(args, db)
    with open('{}.attributes_and_features.pickle'.format(args.database), 'wb') as handle:
        pickle.dump(attributes_and_features, handle,
                    protocol=pickle.HIGHEST_PROTOCOL)  # saving the dictionary
    for feature, attributes in attributes_and_features.items():
        database_struct.write("{}: {}\n".format(feature, ", ".join(set(attributes))))
    
    if args.format == "GTF":
        if (("gene" not in attributes_and_features.keys() and not args.infer_genes) or ("transcript" not in attributes_and_features.keys() and not args.infer_transcripts)):
            eprint("{}ERROR: The file {} seems to be missing gene or transcript features. Please use --infer_genes and/or --infer_transcripts".format(os.linesep, args.annotation))
            sys.exit()
    
    parent_child = set()
    # Find featuretypes without parents
    if args.format == "GFF3":
        for featuretype in db.featuretypes():
            children = set()
            for f in db.iter_by_parent_childs(featuretype, level=1):
                children.update(set([x.featuretype for x in f if x.featuretype != featuretype]))
            if len(children):
                for c in children:
                    parent_child.add((featuretype,c))
    
    if args.format == "GTF":
        parent_child.add(("gene", "transcript"))
        for i in db.featuretypes():
            if i not in ["gene", "transcript"]:
                parent_child.add(("transcript", i))
        
    
    file = open("{}.relations".format(args.database), "w")
    for relation in parent_child:
        file.write("{}\t{}\n".format(relation[0], relation[1]))
    file.close()
    output_lines = []

    database_struct.write(os.linesep)
    database_struct.write(os.linesep)

    parent_child = add_super_feature(parent_child)
   
    imp = importer.IndentedStringImporter()
    try:
        for r in find_roots(parent_child):
            output_lines = []
            find_children(r, parent_child)
            root = imp.import_(output_lines)
            tree = RenderTree(root.children[0])
            for pre, fill, node in tree:
                database_struct.write("%s%s\n" % (pre, node.name))
    except IndexError:
        eprint("{}ERROR: File {} has the wrong format. Couldn't create parent-child relations. If you are using GTF file format, please check if gene and transcript feature lines are present - if not, use --infer_genes option or consider using another annotation file. Check if transcript and gene names differ and are correctly formatted.".format(os.linesep, args.annotation))
        os.system("rm -f {}".format(args.database))
        sys.exit()

    if (not args.save_on_disk):
        bck = sqlite3.connect(args.database)
        with bck:
            db.conn.backup(bck)
        bck.close()
    print("-"*10)
    print("The pipeline finished successfully!")
    print("Available attributes and relations are available in the file {}.database.structure.txt".format(args.database))
    database_struct.close()


if __name__ == "__main__":
    sys.exit(main(sys.argv))



from agouti_pkg.gffutils.create import create_db
from agouti_pkg.gffutils.interface import FeatureDB
from agouti_pkg.gffutils.feature import Feature
from agouti_pkg.gffutils.iterators import DataIterator
from agouti_pkg.gffutils.helpers import example_filename
from agouti_pkg.gffutils.exceptions import FeatureNotFoundError, DuplicateIDError
from agouti_pkg.gffutils.version import version as __version__



from agouti_pkg.miscallaneous import handle_negative_coordinates
from agouti_pkg.processing_product import ProcessingProduct
from agouti_pkg.eprint import eprint
import sys
import codecs
from agouti_pkg.argument_parser import parse_arguments


def read_custom_format(input_file, custom, sep, first_base_num,
                       num_of_bed_fields, header_line_num):
    """Reads and parse input file in CUSTOM format. Returns dictionary of\
        id (key) and ProcessingProduct's (value)

    Arguments:
        input_file {str} -- input file name or path
        custom {str} -- custom file format description - as in --custom arg
        sep {str} -- column separator
        first_base_num {int} -- 0 for 0-based coordinates or 1 for 1-based
        num_of_bed_fields {int} -- number of columns in the input file
        header_line_num {int} -- number of header lines from the top of the file

    Returns:
        tuple -- first element stands for dictionary based on input file.\
                Key stands for product ID, value is an object of\
                ProcessingProduct class. Second element is of int type and\
                describes number of columns in the input file
    """

    products = {}
    custom_format = custom.split(',')
    custom_format = list(filter(None, custom_format))
    try:
        custom_format = list(map(int, custom_format))
    except ValueError:
        eprint("ERROR: incorrect format provided with --custom flag. Invalid literal for int() with base 10")
        sys.exit()
    if (len(custom_format) < 4): 
        eprint("ERROR: incorrect format provided with --custom flag or wrong separator used")
        sys.exit()


    try:
        with open(input_file) as f:
            for index, line in enumerate(f): ###
                if header_line_num - index > 0:
                    if header_line_num - index > 1:
                        print(line.strip())
                    continue
                if not line.startswith("#"):
                    splitted_line = line.split(codecs.decode(
                                                       sep,
                                                       'unicode_escape'))

                    if (num_of_bed_fields == -1):
                        num_of_bed_fields = len(splitted_line)
                    line_updated = ""
                    if line.strip().endswith(sep):
                        line_updated = "{}.\n".format(line.strip())
                    else:
                        line_updated = line
                    

                    start_coord, end_coord, coord_outside_transcript =\
                        handle_negative_coordinates(int(splitted_line[
                                                     custom_format[2] - 1]),
                                                    int(splitted_line[
                                                     custom_format[3] - 1]))

                    coordinates = (splitted_line[custom_format[1] - 1],
                                   start_coord, end_coord)

                    processing_product = splitted_line[custom_format[0] - 1]
                    score = 0

                    if len(custom_format) == 4:
                        strand = "."
                    else:
                        strand = splitted_line[custom_format[4] - 1]

                    if coord_outside_transcript != "No":
                        product = ProcessingProduct(coordinates,
                                                    processing_product,
                                                    score, strand,
                                                    line_updated.strip().replace(sep,
                                                                         "\t"),
                                                    first_base_num,
                                                    coord_outside_transcript)
                    else:
                        product = ProcessingProduct(coordinates,
                                                    processing_product,
                                                    score, strand,
                                                    line_updated.strip().replace(sep,
                                                                         "\t"),
                                                    first_base_num)

                    products[processing_product] = product
        f.close()
    except IndexError:
        eprint("ERROR: incorrect format provided with --custom flag or wrong separator used")
        sys.exit()

    return products, num_of_bed_fields


def read_BED_file(bed_file, num_of_bed_fields, first_base_num, header_line_num):
    """Reads and parse input file in BED format. Returns dictionary of\
        id (key) and ProcessingProduct's (value)

    Arguments:
        bed_file {str} -- input file name or path
        num_of_bed_fields {int} -- number of columns in the input file
        first_base_num {int} -- 0 for 0-based coordinates or 1 for 1-based
        header_line_num {int} -- number of header lines from the top of the file

    Returns:
        [tuple] -- list of ProcessingProducts objects, number of columns in\
                   the input file
    """
    products = {}
    counter = 0
    with open(bed_file) as f: 
        for index, line in enumerate(f): ###
            if header_line_num - index > 0:
                continue
            if not line.startswith("#"):
                tab = line.strip().split()
                num_of_bed_fields = len(list(filter(None, tab))) if (
                    num_of_bed_fields == -1) else num_of_bed_fields
                start_coord, end_coord, coord_outside_transcript = (
                    handle_negative_coordinates(int(tab[1]), int(tab[2])))
                coordinates = (tab[0], start_coord, end_coord)
                while (len(tab) < 6):
                    # in case of less than 6 columns in BED,
                    # fill the rest (up to 6) with blank fields
                    tab.append(".")
                processing_product, score, strand = tab[3], tab[4], tab[5]
                if processing_product == ".":
                    processing_product = f"unnamed_agouti_feature_{counter}"
                    counter += 1
                if coord_outside_transcript != "No":
                    product = ProcessingProduct(coordinates,
                                                processing_product,
                                                score, strand, line.strip(),
                                                first_base_num,
                                                coord_outside_transcript)
                else:
                    product = ProcessingProduct(coordinates,
                                                processing_product,
                                                score, strand, line.strip(),
                                                0)
                products[processing_product] = product
    f.close()
    return products, num_of_bed_fields


def read_header_line(bed_file, custom, sep, header_line_num):
    """Read and parse header line from the input file

    Arguments:
        bed_file {str} -- input file name or path
        custom {str} -- custom file format description - as in --custom arg
        sep {str} -- separator
        header_line_num {int} -- number of header lines from the top of the file

    Returns:
        [tuple] -- header line, list of header fields
    """
    f = open(bed_file)
    lines = f.readlines()
    header = False
    field_list = []
    potential_header = ""
    if header_line_num != 0:
        potential_header = lines[header_line_num - 1]
        header = True
    f.close()
    if custom != "BED" and header:
        custom_format = list(
            map(int, list(filter(None, custom.split(',')))))
        sh = list(filter(None, potential_header.strip().split(codecs.decode(
                                                       sep,
                                                       'unicode_escape'))))
        for i in range(0, len(sh)):
            if (i + 1 not in custom_format):
                field_list.append(sh[i])
    elif custom == "BED":
        sh = list(filter(None, potential_header.strip().split()))
        if len(sh) > 6:
            field_list = sh[6:]
    return header, field_list


# Portions copied over from BCBio.GFF.GFFParser

import re
import copy
import collections
from agouti_pkg.six.moves import urllib
from agouti_pkg.gffutils import constants
from agouti_pkg.gffutils.exceptions import AttributeStringError

import logging

formatter = logging.Formatter('%(asctime)s - %(levelname)s - %(message)s')
logger = logging.getLogger(__name__)
ch = logging.StreamHandler()
ch.setLevel(logging.INFO)
ch.setFormatter(formatter)
logger.addHandler(ch)

gff3_kw_pat = re.compile('\w+=')

# Encoding/decoding notes
# -----------------------
# From
# https://github.com/The-Sequence-Ontology/Specifications/blob/master/gff3.md#description-of-the-format:
#
#       GFF3 files are nine-column, tab-delimited, plain text files.
#       Literal use of tab, newline, carriage return, the percent (%) sign,
#       and control characters must be encoded using RFC 3986
#       Percent-Encoding; no other characters may be encoded. Backslash and
#       other ad-hoc escaping conventions that have been added to the GFF
#       format are not allowed. The file contents may include any character
#       in the set supported by the operating environment, although for
#       portability with other systems, use of Latin-1 or Unicode are
#       recommended.
#
#           tab (%09)
#           newline (%0A)
#           carriage return (%0D)
#           % percent (%25)
#           control characters (%00 through %1F, %7F)
#
#       In addition, the following characters have reserved meanings in
#       column 9 and must be escaped when used in other contexts:
#
#           ; semicolon (%3B)
#           = equals (%3D)
#           & ampersand (%26)
#           , comma (%2C)
#
#
# See also issue #98.
#
# Note that spaces are NOT encoded. Some GFF files have spaces encoded; in
# these cases round-trip invariance will not hold since the %20 will be decoded
# but not re-encoded.
_to_quote = '\n\t\r%;=&,'
_to_quote += ''.join([chr(i) for i in range(32)])
_to_quote += chr(127)


# Caching idea from urllib.parse.Quoter, which uses a defaultdict for
# efficiency. Here we're sort of doing the reverse of the "reserved" idea used
# there.
class Quoter(collections.defaultdict):
    def __missing__(self, b):
        if b in _to_quote:
            res = '%{:02X}'.format(ord(b))
        else:
            res = b
        self[b] = res
        return res

quoter = Quoter()


def _reconstruct(keyvals, dialect, keep_order=False,
                 sort_attribute_values=False):
    """
    Reconstructs the original attributes string according to the dialect.

    Parameters
    ==========
    keyvals : dict
        Attributes from a GFF/GTF feature

    dialect : dict
        Dialect containing info on how to reconstruct a string version of the
        attributes

    keep_order : bool
        If True, then perform sorting of attribute keys to ensure they are in
        the same order as those provided in the original file.  Default is
        False, which saves time especially on large data sets.

    sort_attribute_values : bool
        If True, then sort values to ensure they will always be in the same
        order.  Mostly only useful for testing; default is False.
    """
    if not dialect:
        raise AttributeStringError()
    if not keyvals:
        return ""
    parts = []

    # Re-encode when reconstructing attributes
    if constants.ignore_url_escape_characters or dialect['fmt'] != 'gff3':
        attributes = keyvals
    else:
        attributes = {}
        for k, v in keyvals.items():
            attributes[k] = []
            for i in v:
                attributes[k].append(''.join([quoter[j] for j in i]))

    # May need to split multiple values into multiple key/val pairs
    if dialect['repeated keys']:
        items = []
        for key, val in attributes.items():
            if len(val) > 1:
                for v in val:
                    items.append((key, [v]))
            else:
                items.append((key, val))
    else:
        items = list(attributes.items())

    def sort_key(x):
        # sort keys by their order in the dialect; anything not in there will
        # be in arbitrary order at the end.
        try:
            return dialect['order'].index(x[0])
        except ValueError:
            return 1e6

    if keep_order:
        items.sort(key=sort_key)

    for key, val in items:

        # Multival sep is usually a comma:
        if val:
            if sort_attribute_values:
                val = sorted(val)

            val_str = dialect['multival separator'].join(val)

            if val_str:

                # Surround with quotes if needed
                if dialect['quoted GFF2 values']:
                    val_str = '"%s"' % val_str

                # Typically "=" for GFF3 or " " otherwise
                part = dialect['keyval separator'].join([key, val_str])
        else:
            if dialect['fmt'] == 'gtf':
                part = dialect['keyval separator'].join([key, '""'])
            else:
                part = key
        parts.append(part)

    # Typically ";" or "; "
    parts_str = dialect['field separator'].join(parts)

    # Sometimes need to add this
    if dialect['trailing semicolon']:
        parts_str += ';'

    return parts_str


# TODO:
# Cythonize -- profiling shows that the bulk of the time is spent on this
# function...
def _split_keyvals(keyval_str, dialect=None):
    """
    Given the string attributes field of a GFF-like line, split it into an
    attributes dictionary and a "dialect" dictionary which contains information
    needed to reconstruct the original string.

    Lots of logic here to handle all the corner cases.

    If `dialect` is None, then do all the logic to infer a dialect from this
    attribute string.

    Otherwise, use the provided dialect (and return it at the end).
    """

    def _unquote_quals(quals, dialect):
        """
        Handles the unquoting (decoding) of percent-encoded characters.

        See notes on encoding/decoding above.
        """
        if not constants.ignore_url_escape_characters and dialect['fmt'] == 'gff3':
            for key, vals in quals.items():
                unquoted = [urllib.parse.unquote(v) for v in vals]
                quals[key] = unquoted
        return quals

    infer_dialect = False
    if dialect is None:
        # Make a copy of default dialect so it can be modified as needed
        dialect = copy.copy(constants.dialect)
        infer_dialect = True
    from agouti_pkg.gffutils import feature
    quals = feature.dict_class()
    if not keyval_str:
        return quals, dialect

    # If a dialect was provided, then use that directly.
    if not infer_dialect:
        if dialect['trailing semicolon']:
            keyval_str = keyval_str.rstrip(';')

        parts = keyval_str.split(dialect['field separator'])

        kvsep = dialect['keyval separator']
        if dialect['leading semicolon']:
            pieces = []
            for p in parts:
                if p and p[0] == ';':
                    p = p[1:]
                pieces.append(p.strip().split(kvsep))
                key_vals = [(p[0], " ".join(p[1:])) for p in pieces]

        if dialect['fmt'] == 'gff3':
            key_vals = [p.split(kvsep) for p in parts]
        else:
            leadingsemicolon = dialect['leading semicolon']
            pieces = []
            for i, p in enumerate(parts):
                if i == 0 and leadingsemicolon:
                    p = p[1:]
                pieces.append(p.strip().split(kvsep))
                key_vals = [(p[0], " ".join(p[1:])) for p in pieces]

        quoted = dialect['quoted GFF2 values']
        for item in key_vals:
            # Easy if it follows spec
            if len(item) == 2:
                key, val = item

            # Only key provided?
            elif len(item) == 1:
                key = item[0]
                val = ''

            else:
                key = item[0]
                val = dialect['keyval separator'].join(item[1:])

            try:
                quals[key]
            except KeyError:
                quals[key] = []

            if quoted:
                if (len(val) > 0 and val[0] == '"' and val[-1] == '"'):
                    val = val[1:-1]

            if val:
                # TODO: if there are extra commas for a value, just use empty
                # strings
                # quals[key].extend([v for v in val.split(',') if v])
                vals = val.split(',')
                quals[key].extend(vals)

        quals = _unquote_quals(quals, dialect)
        return quals, dialect

    # If we got here, then we need to infer the dialect....
    #
    # Reset the order to an empty list so that it will only be populated with
    # keys that are found in the file.
    dialect['order'] = []

    # ensembl GTF has trailing semicolon
    if keyval_str[-1] == ';':
        keyval_str = keyval_str[:-1]
        dialect['trailing semicolon'] = True

    # GFF2/GTF has a semicolon with at least one space after it.
    # Spaces can be on both sides (e.g. wormbase)
    # GFF3 works with no spaces.
    # So split on the first one we can recognize...
    for sep in (' ; ', '; ', ';'):
        parts = keyval_str.split(sep)
        if len(parts) > 1:
            dialect['field separator'] = sep
            break

    # Is it GFF3?  They have key-vals separated by "="
    if gff3_kw_pat.match(parts[0]):
        key_vals = [p.split('=') for p in parts]
        dialect['fmt'] = 'gff3'
        dialect['keyval separator'] = '='

    # Otherwise, key-vals separated by space.  Key is first item.
    else:
        dialect['keyval separator'] = " "
        pieces = []
        for p in parts:
            # Fix misplaced semicolons in keys in some GFF2 files
            if p and p[0] == ';':
                p = p[1:]
                dialect['leading semicolon'] = True
            pieces.append(p.strip().split(' '))
        key_vals = [(p[0], " ".join(p[1:])) for p in pieces]

    for item in key_vals:

        # Easy if it follows spec
        if len(item) == 2:
            key, val = item

        # Only key provided?
        elif len(item) == 1:
                key = item[0]
                val = ''

        # Pathological cases where values of a key have within them the key-val
        # separator, e.g.,
        #  Alias=SGN-M1347;ID=T0028;Note=marker name(s): T0028 SGN-M1347 |identity=99.58|escore=2e-126
        else:
            key = item[0]
            val = dialect['keyval separator'].join(item[1:])

        # Is the key already in there?
        if key in quals:
            dialect['repeated keys'] = True
        else:
            quals[key] = []

        # Remove quotes in GFF2
        if len(val) > 0 and val[0] == '"' and val[-1] == '"':
            val = val[1:-1]
            dialect['quoted GFF2 values'] = True
        if val:
            # TODO: if there are extra commas for a value, just use empty
            # strings
            # quals[key].extend([v for v in val.split(',') if v])
            vals = val.split(',')
            if (len(vals) > 1) and dialect['repeated keys']:
                raise AttributeStringError(
                    "Internally inconsistent attributes formatting: "
                    "some have repeated keys, some do not.")
            quals[key].extend(vals)

        # keep track of the order of keys
        dialect['order'].append(key)

    if (
        (dialect['keyval separator'] == ' ') and
        (dialect['quoted GFF2 values'])
    ):
        dialect['fmt'] = 'gtf'

    quals = _unquote_quals(quals, dialect)
    return quals, dialect



import copy
import sys
import os
import agouti_pkg.simplejson
import time
import tempfile
import agouti_pkg.six
from agouti_pkg.gffutils import constants
from agouti_pkg.gffutils import bins
import agouti_pkg.gffutils
from agouti_pkg.gffutils import gffwriter
from agouti_pkg.gffutils import parser
from agouti_pkg.gffutils.attributes import dict_class

HERE = os.path.dirname(os.path.abspath(__file__))


def example_filename(fn):
    """
    Return the full path of a data file that ships with gffutils.
    """
    return os.path.join(HERE, 'test', 'data', fn)


def infer_dialect(attributes):
    """
    Infer the dialect based on the attributes.

    Parameters
    ----------
    attributes : str or iterable
        A single attributes string from a GTF or GFF line, or an iterable of
        such strings.

    Returns
    -------
    Dictionary representing the inferred dialect
    """
    if isinstance(attributes, agouti_pkg.six.string_types):
        attributes = [attributes]
    dialects = [parser._split_keyvals(i)[1] for i in attributes]
    return _choose_dialect(dialects)


def _choose_dialect(dialects):
    """
    Given a list of dialects, choose the one to use as the "canonical" version.

    If `dialects` is an empty list, then use the default GFF3 dialect

    Parameters
    ----------
    dialects : iterable
        iterable of dialect dictionaries

    Returns
    -------
    dict
    """
    # NOTE: can use helpers.dialect_compare if you need to make this more
    # complex....

    # For now, this function favors the first dialect, and then appends the
    # order of additional fields seen in the attributes of other lines giving
    # priority to dialects that come first in the iterable.
    if len(dialects) == 0:
        return constants.dialect
    final_order = []
    for dialect in dialects:
        for o in dialect['order']:
            if o not in final_order:
                final_order.append(o)
    dialect = dialects[0]
    dialect['order'] = final_order
    return dialect


def make_query(args, other=None, limit=None, strand=None, featuretype=None,
               extra=None, order_by=None, reverse=False,
               completely_within=False):
    """
    Multi-purpose, bare-bones ORM function.

    This function composes queries given some commonly-used kwargs that can be
    passed to FeatureDB methods (like .parents(), .children(), .all_features(),
    .features_of_type()).  It handles, in one place, things like restricting to
    featuretype, limiting to a genomic range, limiting to one strand, or
    returning results ordered by different criteria.

    Additional filtering/subsetting/sorting behavior should be added here.

    (Note: this ended up having better performance (and flexibility) than
    sqlalchemy)

    This function also provides support for additional JOINs etc (supplied via
    the `other` kwarg) and extra conditional clauses (`extra` kwarg).  See the
    `_QUERY` var below for the order in which they are used.

    For example, FeatureDB._relation uses `other` to supply the JOIN
    substatment, and that same method also uses `extra` to supply the
    "relations.level = ?" substatment (see the source for FeatureDB._relation
    for more details).

    `args` contains the arguments that will ultimately be supplied to the
    sqlite3.connection.execute function.  It may be further populated below --
    for example, if strand="+", then the query will include a strand clause,
    and the strand will be appended to the args.

    `args` can be pre-filled with args that are passed to `other` and `extra`.
    """

    _QUERY = ("{_SELECT} {OTHER} {EXTRA} {FEATURETYPE} "
              "{LIMIT} {STRAND} {ORDER_BY}")

    # Construct a dictionary `d` that will be used later as _QUERY.format(**d).
    # Default is just _SELECT, which returns all records in the features table.
    # (Recall that constants._SELECT gets the fields in the order needed to
    # reconstruct a Feature)
    d = dict(_SELECT=constants._SELECT, OTHER="", FEATURETYPE="", LIMIT="",
             STRAND="", ORDER_BY="", EXTRA="")

    if other:
        d['OTHER'] = other
    if extra:
        d['EXTRA'] = extra

    # If `other` and `extra` take args (that is, they have "?" in them), then
    # they should have been provided in `args`.
    required_args = (d['EXTRA'] + d['OTHER']).count('?')
    if len(args) != required_args:
        raise ValueError('Not enough args (%s) for subquery' % args)

    # Below, if a kwarg is specified, then we create sections of the query --
    # appending to args as necessary.
    #
    # IMPORTANT: the order in which things are processed here is the same as
    # the order of the placeholders in _QUERY.  That is, we need to build the
    # args in parallel with the query to avoid putting the wrong args in the
    # wrong place.

    if featuretype:
        # Handle single or iterables of featuretypes.
        #
        # e.g., "featuretype = 'exon'"
        #
        # or, "featuretype IN ('exon', 'CDS')"
        if isinstance(featuretype, agouti_pkg.six.string_types):
            d['FEATURETYPE'] = "features.featuretype = ?"
            args.append(featuretype)
        else:
            d['FEATURETYPE'] = (
                "features.featuretype IN  (%s)"
                % (','.join(["?" for _ in featuretype]))
            )
            args.extend(featuretype)

    if limit:
        # Restrict to a genomic region.  Makes use of the UCSC binning strategy
        # for performance.
        #
        # `limit` is a string or a tuple of (chrom, start, stop)
        #
        # e.g., "seqid = 'chr2L' AND start > 1000 AND end < 5000"
        if isinstance(limit, agouti_pkg.six.string_types):
            seqid, startstop = limit.split(':')
            start, end = startstop.split('-')
        else:
            seqid, start, end = limit

        # Identify possible bins
        _bins = bins.bins(int(start), int(end), one=False)

        # Use different overlap conditions
        if completely_within:
            d['LIMIT'] = (
                "features.seqid = ? AND features.start >= ? "
                "AND features.end <= ?"
            )
            args.extend([seqid, start, end])

        else:
            d['LIMIT'] = (
                "features.seqid = ? AND features.start <= ? "
                "AND features.end >= ?"
            )
            # Note order (end, start)
            args.extend([seqid, end, start])

        # Add bin clause. See issue #45.
        if len(_bins) < 900:
            d['LIMIT'] += " AND features.bin IN (%s)" % (','.join(map(str, _bins)))

    if strand:
        # e.g., "strand = '+'"
        d['STRAND'] = "features.strand = ?"
        args.append(strand)

    # TODO: implement file_order!
    valid_order_by = constants._gffkeys_extra + ['file_order', 'length']
    _order_by = []
    if order_by:
        # Default is essentially random order.
        #
        # e.g. "ORDER BY seqid, start DESC"
        if isinstance(order_by, agouti_pkg.six.string_types):
            _order_by.append(order_by)

        else:
            for k in order_by:
                if k not in valid_order_by:
                    raise ValueError("%s not a valid order-by value in %s"
                                     % (k, valid_order_by))

                # There's no length field, so order by end - start
                if k == 'length':
                    k = '(end - start)'

                _order_by.append(k)

        _order_by = ','.join(_order_by)
        if reverse:
            direction = 'DESC'
        else:
            direction = 'ASC'
        d['ORDER_BY'] = 'ORDER BY %s %s' % (_order_by, direction)

    # Ensure only one "WHERE" is included; the rest get "AND ".  This is ugly.
    where = False
    if "where" in d['OTHER'].lower():
        where = True
    for i in ['EXTRA', 'FEATURETYPE', 'LIMIT', 'STRAND']:
        if d[i]:
            if not where:
                d[i] = "WHERE " + d[i]
                where = True
            else:
                d[i] = "AND " + d[i]

    return _QUERY.format(**d), args


def _bin_from_dict(d):
    """
    Given a dictionary yielded by the parser, return the genomic "UCSC" bin
    """
    try:
        start = int(d['start'])
        end = int(d['end'])
        return bins.bins(start, end, one=True)

    # e.g., if "."
    except ValueError:
        return None


def _jsonify(x):
    """Use most compact form of JSON"""
    if isinstance(x, dict_class):
        return agouti_pkg.simplejson.dumps(x._d, separators=(',', ':'))
    return agouti_pkg.simplejson.dumps(x, separators=(',', ':'))


def _unjsonify(x, isattributes=False):
    """Convert JSON string to an ordered defaultdict."""
    if isattributes:
        obj = agouti_pkg.simplejson.loads(x)
        return dict_class(obj)
    return agouti_pkg.simplejson.loads(x)


def _feature_to_fields(f, jsonify=True):
    """
    Convert feature to tuple, for faster sqlite3 import
    """
    x = []
    for k in constants._keys:
        v = getattr(f, k)
        if jsonify and (k in ('attributes', 'extra')):
            x.append(_jsonify(v))
        else:
            x.append(v)
    return tuple(x)


def _dict_to_fields(d, jsonify=True):
    """
    Convert dict to tuple, for faster sqlite3 import
    """
    x = []
    for k in constants._keys:
        v = d[k]
        if jsonify and (k in ('attributes', 'extra')):
            x.append(_jsonify(v))
        else:
            x.append(v)
    return tuple(x)


def asinterval(feature):
    """
    Converts a gffutils.Feature to a pybedtools.Interval
    """
    import pybedtools
    return pybedtools.create_interval_from_list(str(feature).split('\t'))


def merge_attributes(attr1, attr2):
    """
    Merges two attribute dictionaries into a single dictionary.

    Parameters
    ----------
    `attr1`, `attr2` : dict

    Returns
    -------
    dict
    """

    new_d = copy.deepcopy(attr1)
    new_d.update(attr2)

    #all of attr2 key : values just overwrote attr1, fix it
    for k, v in new_d.items():
        if not isinstance(v, list):
            new_d[k] = [v]

    for k, v in agouti_pkg.six.iteritems(attr1):
        if k in attr2:
            if not isinstance(v, list):
                v = [v]
            new_d[k].extend(v)
    return dict((k, list(set(v))) for k, v in new_d.items())


def dialect_compare(dialect1, dialect2):
    """
    Compares two dialects.
    """
    orig = set(dialect1.items())
    new = set(dialect2.items())
    return dict(
        added=dict(list(new.difference(orig))),
        removed=dict(list(orig.difference(new)))
    )


def sanitize_gff_db(db, gid_field="gid"):
    """
    Sanitize given GFF db. Returns a sanitized GFF db.

    Sanitizing means:

    - Ensuring that start < stop for all features
    - Standardizing gene units by adding a 'gid' attribute
      that makes the file grep-able

    TODO: Do something with negative coordinates?
    """
    def sanitized_iterator():
        # Iterate through the database by each gene's records
        for gene_recs in db.iter_by_parent_childs():
            # The gene's ID
            gene_id = gene_recs[0].id
            for rec in gene_recs:
                # Fixup coordinates if necessary
                if rec.start > rec.stop:
                    rec.start, rec.stop = rec.stop, rec.start
                # Add a gene id field to each gene's records
                rec.attributes[gid_field] = [gene_id]
                yield rec
    # Return sanitized GFF database
    sanitized_db = \
        agouti_pkg.gffutils.create_db(sanitized_iterator(), ":memory:",
                           verbose=False)
    return sanitized_db


def sanitize_gff_file(gff_fname,
                      in_memory=True,
                      in_place=False):
    """
    Sanitize a GFF file.
    """
    db = None
    if is_gff_db(gff_fname):
        # It's a database filename, so load it
        db = agouti_pkg.gffutils.FeatureDB(gff_fname)
    else:
        # Need to create a database for file
        if in_memory:
            db = agouti_pkg.gffutils.create_db(gff_fname, ":memory:",
                                    verbose=False)
        else:
            db = get_gff_db(gff_fname)
    if in_place:
        gff_out = gffwriter.GFFWriter(gff_fname,
                                      in_place=in_place)
    else:
        gff_out = gffwriter.GFFWriter(sys.stdout)
    sanitized_db = sanitize_gff_db(db)
    for gene_rec in sanitized_db.all_features(featuretype="gene"):
        gff_out.write_gene_recs(sanitized_db, gene_rec.id)
    gff_out.close()


def annotate_gff_db(db):
    """
    Annotate a GFF file by cross-referencing it with another GFF
    file, e.g. one containing gene models.
    """
    pass


def is_gff_db(db_fname):
    """
    Return True if the given filename is a GFF database.

    For now, rely on .db extension.
    """
    if not os.path.isfile(db_fname):
        return False
    if db_fname.endswith(".db"):
        return True
    return False


def to_unicode(obj, encoding='utf-8'):
    if isinstance(obj, agouti_pkg.six.string_types):
        if not isinstance(obj, agouti_pkg.six.text_type):
            obj = agouti_pkg.six.text_type(obj, encoding)
    return obj


def canonical_transcripts(db, fasta_filename):
    import pyfaidx
    fasta = pyfaidx.Fasta(fasta_filename, as_raw=True)
    for gene in db.features_of_type('gene'):

        # exons_list will contain (CDS_length, total_length, transcript, [exons]) tuples.
        exon_list = []
        for ti, transcript in enumerate(db.children(gene, level=1)):
            cds_len = 0
            total_len = 0
            exons = list(db.children(transcript, level=1))
            for exon in exons:
                exon_length = len(exon)
                if exon.featuretype == 'CDS':
                    cds_len += exon_length
                total_len += exon_length

            exon_list.append((cds_len, total_len, transcript, exons))

        # If we have CDS, then use the longest coding transcript
        if max(i[0] for i in exon_list) > 0:
            best = sorted(exon_list)[0]
        # Otherwise, just choose the longest
        else:
            best = sorted(exon_list, lambda x: x[1])[0]

        print(best)

        canonical_exons = best[-1]
        transcript = best[-2]
        seqs = [i.sequence(fasta) for i in canonical_exons]
        yield transcript, ''.join(seqs)



##
## Helpers for gffutils-cli
##
## TODO: move clean_gff here?
##
def get_gff_db(gff_fname,
               ext=".db"):
    """
    Get db for GFF file. If the database has a .db file,
    load that. Otherwise, create a named temporary file,
    serialize the db to that, and return the loaded database.
    """
    if not os.path.isfile(gff_fname):
        # Not sure how we should deal with errors normally in
        # gffutils -- Ryan?
        raise ValueError("GFF %s does not exist." % (gff_fname))
    candidate_db_fname = "%s.%s" % (gff_fname, ext)
    if os.path.isfile(candidate_db_fname):
        # Standard .db file found, so return it
        return candidate_db_fname
    # Otherwise, we need to create a temporary but non-deleted
    # file to store the db in. It'll be up to the user
    # of the function the delete the file when done.
    ## NOTE: Ryan must have a good scheme for dealing with this
    ## since pybedtools does something similar under the hood, i.e.
    ## creating temporary files as needed without over proliferation
    db_fname = tempfile.NamedTemporaryFile(delete=False)
    # Create the database for the gff file (suppress output
    # when using function internally)
    print("Creating db for %s" % (gff_fname))
    t1 = time.time()
    db = agouti_pkg.gffutils.create_db(gff_fname, db_fname.name,
                            merge_strategy="merge",
                            verbose=False)
    t2 = time.time()
    print("  - Took %.2f seconds" % (t2 - t1))
    return db



import tempfile
from nose.tools import assert_raises
from agouti.gffutils import (parser, create, feature, iterators, constants, helpers,
                      exceptions)
from agouti.gffutils import example_filename, create_db
from . import attr_test_cases
from textwrap import dedent

TEST_FILENAMES = [example_filename(i) for i in [
    'c_elegans_WS199_ann_gff.txt',
    'ensembl_gtf.txt',
    'hybrid1.gff3',
    'ncbi_gff3.txt',
    'c_elegans_WS199_dna_shortened.fa',
    'F3-unique-3.v2.gff',
    'jgi_gff2.txt',
    'wormbase_gff2_alt.txt',
    'c_elegans_WS199_shortened_gff.txt',
    'glimmer_nokeyval.gff3',
    'mouse_extra_comma.gff3',
    'wormbase_gff2.txt']]


def test_directives():
    data = dedent("""
    ##directive1 example
    .	.	.	.	.	.	.	.
    .	.	.	.	.	.	.	.
    .	.	.	.	.	.	.	.
    .	.	.	.	.	.	.	.
    """)

    it = iterators._StringIterator(data)
    db = create_db(data, dbfn=':memory:', from_string=True, verbose=False)
    assert it.directives == db.directives == ['directive1 example'], (it.directives, db.directives)

def test_split_attrs():
    # nosetests generator for all the test cases in attr_test_cases.  (note no
    # docstring for this test function so that nosetests -v will print the test
    # cases)
    for (attr_str, attr_dict, acceptable_reconstruction) \
            in attr_test_cases.attrs:
        yield attrs_OK, attr_str, attr_dict, acceptable_reconstruction


def attrs_OK(attr_str, attr_dict, acceptable_reconstruction=None):
    """
    Given an attribute string and a dictionary of what you expect, test the
    attribute splitting and reconstruction (invariant roundtrip).

    There are some corner cases for the roundtrip invariance that don't work
    (see attr_test_cases.py for details); `acceptable_reconstruction` handles
    those.
    """
    result, dialect = parser._split_keyvals(attr_str)
    assert result == attr_dict, result

    reconstructed = parser._reconstruct(result, dialect, keep_order=True)
    if acceptable_reconstruction:
        assert reconstructed == acceptable_reconstruction, reconstructed
    else:
        assert reconstructed == attr_str, reconstructed


def parser_smoke_test():
    """
    Just confirm we can iterate completely through the test files....
    """
    # Don't show the warnings for tests
    import logging
    parser.logger.setLevel(logging.CRITICAL)
    for filename in TEST_FILENAMES:
        p = iterators._FileIterator(filename)
        for i in p:
            continue

def test_empty_recontruct():
    """
    reconstructing attributes with incomplete information returns empty string
    """
    assert parser._reconstruct(None, constants.dialect) == ""
    assert_raises(exceptions.AttributeStringError, parser._reconstruct, dict(ID='asdf'), None)
    assert_raises(exceptions.AttributeStringError, parser._reconstruct, None, None)

def test_empty_split_keyvals():
    attrs, dialect = parser._split_keyvals(keyval_str=None)
    assert attrs == feature.dict_class()
    assert dialect == constants.dialect

def test_repeated_keys_conflict():
    """
    if dialect says repeated keys, but len(vals) > 1, then the keys are not
    actually repeated....
    """
    #
    # This is now only checked when infer_dialect is True -- and is disabled
    # when a dialect is provided
    #
    #dialect = constants.dialect.copy()
    #dialect['repeated keys'] = True
    #assert_raises(exceptions.AttributeStringError, parser._split_keyvals, "Parent=1,2,3", dialect)

def test_parser_from_string():
    """
    make sure from string and from file return identical results
    """
    line = b"chr2L	FlyBase	exon	7529	8116	.	+	.	Name=CG11023:1;Parent=FBtr0300689,FBtr0300690"
    tmp = tempfile.NamedTemporaryFile()
    tmp.write(line)
    tmp.seek(0)

    p1 = iterators._StringIterator(
        "chr2L	FlyBase	exon	7529	8116	.	+	.	Name=CG11023:1;Parent=FBtr0300689,FBtr0300690"
    )
    p2 = iterators._FileIterator(tmp.name)
    lines = list(zip(p1, p2))
    assert len(lines) == 1
    assert p1.current_item_number == p2.current_item_number == 0
    assert lines[0][0] == lines[0][1]


def test_valid_line_count():
    p = iterators._FileIterator(example_filename('ncbi_gff3.txt'))
    assert len(list(p)) == 17

    p = iterators._FileIterator(example_filename('hybrid1.gff3'))
    assert len(list(p)) == 6

    p = iterators._FileIterator(example_filename('FBgn0031208.gff'))
    assert len(list(p)) == 27

def test_inconsistent_dialect():
    """
    The second feature does not have a trailing semicolon (wormbase_gff2_alt is
    like this).  But since the first feature does, that's what the db's dialect
    is set to, which can cause errors when parsing attributes.
    """
    db = create.create_db(
    """
    chr1	.	gene	1	100	.	+	.	gene_id "gene1";
    chr1	.	mRNA	1	100	.	+	.	transcript_id "mRNA1"
    """, ':memory:', from_string=True)
    items = list(db.all_features())
    print(items[0])
    # before, was ['"mRNA1'] -- note extra "
    assert items[1].attributes['transcript_id'] == ['mRNA1'], items[1].attributes['transcript_id']

def test_attributes():
    s = "chr2L	FlyBase	mRNA	7529	9484	.	+	.	ID=FBtr0300690;Name=CG11023-RC;Parent=FBgn0031208;"
    f = feature.feature_from_line(s)
    f.keep_order = True
    assert str(f) == s, str(f)




#!/usr/bin/env python
# coding=utf-8

import os
import sys
import codecs
from agouti_pkg.eprint import eprint
import agouti_pkg.gffutils
import agouti_pkg.gffutils.inspect as inspect
from operator import attrgetter
from agouti_pkg.database import Database
from agouti_pkg.argument_parser import (parse_arguments,
                             create_attributes_and_features_dict)
from agouti_pkg.sequence_ontology import *
from agouti_pkg.processing_product import ProcessingProduct
from agouti_pkg.miscallaneous import *
from agouti_pkg.read_input import *
from agouti_pkg.header import *
from agouti_pkg.output_processing import prepare_output


def main(args):
    """Main function

    Arguments:
        args {argparse.Namespace} -- command line arguments parsed with\
            argparse
    """

    chromosomes_not_found = set()  ## chromosomes in the input file but not found in the reference annotations
    chromosomes_found = set()  ## chromosomes in the input file and found in the reference annotations

    num_of_bed_fields = -1
    lengths_dict = {}  # stores length of UTRs and CDS of a given transcript

    try:
        db = Database(agouti_pkg.gffutils.FeatureDB(args.database, keep_order=False),
                    args.database)  # reading the database from file
    except ValueError:
        eprint("ERROR: the database provided does not exists")
        sys.exit()

    # list all featuretypes existing in db
    featuretypes_from_db = list(db.database.featuretypes())

    try:

        attr_and_feat = create_attributes_and_features_dict  # abbr
        attributes_and_features = attr_and_feat(db, args, featuretypes_from_db)


    except IndexError:

        eprint("ERROR: arguments provided with --combine, --select_attributes or --select_features are incorrect")
        sys.exit()

    try:
        if (args.custom == "BED"):
            try:
                products, num_of_bed_fields = read_BED_file(args.bed,
                                                            num_of_bed_fields,
                                                            args.first_base_num, args.header_lines)
            except FileNotFoundError:
                eprint("ERROR: the input file does not exists")
                sys.exit()
        else:
            try:
                products, num_of_bed_fields = read_custom_format(args.bed, args.custom,
                                                                args.sep,
                                                                args.first_base_num,
                                                                num_of_bed_fields, args.header_lines)
            except FileNotFoundError:
                eprint("ERROR: the input file does not exists")
                sys.exit()
    except IndexError:
        eprint("ERROR: the input file has the wrong format")
        sys.exit()

    header = prepare_header(db, attributes_and_features, args,
                            num_of_bed_fields, args.header_lines)
    
    whole_output = f"{header[0].strip()}"

    for key, value in products.items():
        if (args.transcriptomic):
            try:
                id = value.coordinates[0]

                overlapping_features = [db.database[id]]
                g2t, cds_start = value.genomic_to_transcriptomic(
                    db, featuretypes_from_db)
                lengths_dict[value.coordinates[0]] = g2t

                out = prepare_output(lengths_dict, header, args,
                                     attributes_and_features, db, value,
                                     overlapping_features, g2t, cds_start)
                
                whole_output = f"{whole_output}\n{out}"

            except (agouti_pkg.gffutils.exceptions.FeatureNotFoundError):
                eprint("WARNING: Couldn't find transcript {}. Please make sure that it exists in your GTF/GFF3 file.".format(id))

        else:

            region = value.coordinates

            ## test whether chromosome is present in annotations
            if value.coordinates[0] not in chromosomes_found:
                if len(list(db.database.region(seqid=value.coordinates[0]))) == 0:
                    chromosomes_not_found.add(value.coordinates[0])
                else:
                    chromosomes_found.add(value.coordinates[0])
            ##
            
            if (not args.strand_specific):
                overlapping_features = list(
                    db.database.region(region=region,
                                       featuretype=attributes_and_features.keys(),
                                       completely_within=False))

            elif(args.strand_specific):
                overlapping_features = list(db.database.region(region=region,
                                                               strand=value.strand,
                                                               featuretype=attributes_and_features.keys(),
                                                               completely_within=False))

            out = prepare_output(lengths_dict, header, args,
                                 attributes_and_features, db, value,
                                 overlapping_features, region)
            
            whole_output = f"{whole_output}\n{out}"

    if args.statistics or args.stats_only:
        statistics(whole_output)
    
    if not args.stats_only:
        print(whole_output)

    if len(chromosomes_not_found):
        eprint(f"WARNING: The following chromosomes were not found in the annotations: {', '.join(list(chromosomes_not_found))}")
    return


if __name__ == "__main__":
    sys.exit(main(sys.argv))



"""
Module for integration with BioPython, specifically SeqRecords and SeqFeature
objects.
"""
import agouti_pkg.six
try:
    from Bio.SeqFeature import SeqFeature, FeatureLocation
    from Bio import SeqIO
except ImportError:
    import warnings
    warnings.warn("BioPython must be installed to use this module")
from .feature import Feature, feature_from_line

_biopython_strand = {
    '+':  1,
    '-': -1,
    '.':  0,
}
_feature_strand = dict((v, k) for k, v in _biopython_strand.items())


def to_seqfeature(feature):
    """
    Converts a gffutils.Feature object to a Bio.SeqFeature object.

    The GFF fields `source`, `score`, `seqid`, and `frame` are stored as
    qualifiers.  GFF `attributes` are also stored as qualifiers.

    Parameters
    ----------
    feature : Feature object, or string
        If string, assume it is a GFF or GTF-format line; otherwise just use
        the provided feature directly.
    """
    if isinstance(feature, agouti_pkg.six.string_types):
        feature = feature_from_line(feature)

    qualifiers = {
        'source': [feature.source],
        'score': [feature.score],
        'seqid': [feature.seqid],
        'frame': [feature.frame],
    }
    qualifiers.update(feature.attributes)
    return SeqFeature(
        # Convert from GFF 1-based to standard Python 0-based indexing used by
        # BioPython
        FeatureLocation(feature.start - 1, feature.stop),
        id=feature.id,
        type=feature.featuretype,
        strand=_biopython_strand[feature.strand],
        qualifiers=qualifiers
    )


def from_seqfeature(s, **kwargs):
    """
    Converts a Bio.SeqFeature object to a gffutils.Feature object.

    The GFF fields `source`, `score`, `seqid`, and `frame` are assumed to be
    stored as qualifiers.  Any other qualifiers will be assumed to be GFF
    attributes.
    """
    source = s.qualifiers.get('source', '.')[0]
    score = s.qualifiers.get('score', '.')[0]
    seqid = s.qualifiers.get('seqid', '.')[0]
    frame = s.qualifiers.get('frame', '.')[0]
    strand = _feature_strand[s.strand]

    # BioPython parses 1-based GenBank positions into 0-based for use within
    # Python.  We need to convert back to 1-based GFF format here.
    start = s.location.start.position + 1
    stop = s.location.end.position
    featuretype = s.type
    id = s.id
    attributes = dict(s.qualifiers)
    attributes.pop('source', '.')
    attributes.pop('score', '.')
    attributes.pop('seqid', '.')
    attributes.pop('frame', '.')
    return Feature(seqid, source, featuretype, start, stop, score, strand,
                   frame, attributes, **kwargs)



# -*- coding: utf-8 -*-

import warnings

from agouti_pkg.anytree.iterators import PreOrderIter

from .exceptions import LoopError
from .exceptions import TreeError


class NodeMixin(object):

    __slots__ = ("__parent", "__children")

    separator = "/"

    u"""
    The :any:`NodeMixin` class extends any Python class to a tree node.

    The only tree relevant information is the `parent` attribute.
    If `None` the :any:`NodeMixin` is root node.
    If set to another node, the :any:`NodeMixin` becomes the child of it.

    >>> from anytree import NodeMixin, RenderTree
    >>> class MyBaseClass(object):
    ...     foo = 4
    >>> class MyClass(MyBaseClass, NodeMixin):  # Add Node feature
    ...     def __init__(self, name, length, width, parent=None):
    ...         super(MyClass, self).__init__()
    ...         self.name = name
    ...         self.length = length
    ...         self.width = width
    ...         self.parent = parent

    >>> my0 = MyClass('my0', 0, 0)
    >>> my1 = MyClass('my1', 1, 0, parent=my0)
    >>> my2 = MyClass('my2', 0, 2, parent=my0)

    >>> for pre, _, node in RenderTree(my0):
    ...     treestr = u"%s%s" % (pre, node.name)
    ...     print(treestr.ljust(8), node.length, node.width)
    my0      0 0
    ├── my1  1 0
    └── my2  0 2
    """

    @property
    def parent(self):
        u"""
        Parent Node.

        On set, the node is detached from any previous parent node and attached
        to the new node.

        >>> from anytree import Node, RenderTree
        >>> udo = Node("Udo")
        >>> marc = Node("Marc")
        >>> lian = Node("Lian", parent=marc)
        >>> print(RenderTree(udo))
        Node('/Udo')
        >>> print(RenderTree(marc))
        Node('/Marc')
        └── Node('/Marc/Lian')

        **Attach**

        >>> marc.parent = udo
        >>> print(RenderTree(udo))
        Node('/Udo')
        └── Node('/Udo/Marc')
            └── Node('/Udo/Marc/Lian')

        **Detach**

        To make a node to a root node, just set this attribute to `None`.

        >>> marc.is_root
        False
        >>> marc.parent = None
        >>> marc.is_root
        True
        """
        try:
            return self.__parent
        except AttributeError:
            return None

    @parent.setter
    def parent(self, value):
        if value is not None and not isinstance(value, NodeMixin):
            msg = "Parent node %r is not of type 'NodeMixin'." % (value)
            raise TreeError(msg)
        try:
            parent = self.__parent
        except AttributeError:
            parent = None
        if parent is not value:
            self.__check_loop(value)
            self.__detach(parent)
            self.__attach(value)

    def __check_loop(self, node):
        if node is not None:
            if node is self:
                msg = "Cannot set parent. %r cannot be parent of itself."
                raise LoopError(msg % self)
            if self in node.path:
                msg = "Cannot set parent. %r is parent of %r."
                raise LoopError(msg % (self, node))

    def __detach(self, parent):
        if parent is not None:
            self._pre_detach(parent)
            parentchildren = parent.__children_
            assert any([child is self for child in parentchildren]), "Tree internal data is corrupt."
            # ATOMIC START
            parentchildren.remove(self)
            self.__parent = None
            # ATOMIC END
            self._post_detach(parent)

    def __attach(self, parent):
        if parent is not None:
            self._pre_attach(parent)
            parentchildren = parent.__children_
            assert not any([child is self for child in parentchildren]), "Tree internal data is corrupt."
            # ATOMIC START
            parentchildren.append(self)
            self.__parent = parent
            # ATOMIC END
            self._post_attach(parent)

    @property
    def __children_(self):
        try:
            return self.__children
        except AttributeError:
            self.__children = []
            return self.__children

    @property
    def children(self):
        """
        All child nodes.

        >>> from anytree import Node
        >>> n = Node("n")
        >>> a = Node("a", parent=n)
        >>> b = Node("b", parent=n)
        >>> c = Node("c", parent=n)
        >>> n.children
        (Node('/n/a'), Node('/n/b'), Node('/n/c'))

        Modifying the children attribute modifies the tree.

        **Detach**

        The children attribute can be updated by setting to an iterable.

        >>> n.children = [a, b]
        >>> n.children
        (Node('/n/a'), Node('/n/b'))

        Node `c` is removed from the tree.
        In case of an existing reference, the node `c` does not vanish and is the root of its own tree.

        >>> c
        Node('/c')

        **Attach**

        >>> d = Node("d")
        >>> d
        Node('/d')
        >>> n.children = [a, b, d]
        >>> n.children
        (Node('/n/a'), Node('/n/b'), Node('/n/d'))
        >>> d
        Node('/n/d')

        **Duplicate**

        A node can just be the children once. Duplicates cause a :any:`TreeError`:

        >>> n.children = [a, b, d, a]
        Traceback (most recent call last):
            ...
        anytree.node.exceptions.TreeError: Cannot add node Node('/n/a') multiple times as child.
        """
        return tuple(self.__children_)

    @staticmethod
    def __check_children(children):
        seen = set()
        for child in children:
            if not isinstance(child, NodeMixin):
                msg = ("Cannot add non-node object %r. "
                       "It is not a subclass of 'NodeMixin'.") % child
                raise TreeError(msg)
            if child not in seen:
                seen.add(child)
            else:
                msg = "Cannot add node %r multiple times as child." % child
                raise TreeError(msg)

    @children.setter
    def children(self, children):
        # convert iterable to tuple
        children = tuple(children)
        NodeMixin.__check_children(children)
        # ATOMIC start
        old_children = self.children
        del self.children
        try:
            self._pre_attach_children(children)
            for child in children:
                child.parent = self
            self._post_attach_children(children)
            assert len(self.children) == len(children)
        except Exception:
            self.children = old_children
            raise
        # ATOMIC end

    @children.deleter
    def children(self):
        children = self.children
        self._pre_detach_children(children)
        for child in self.children:
            child.parent = None
        assert len(self.children) == 0
        self._post_detach_children(children)

    def _pre_detach_children(self, children):
        """Method call before detaching `children`."""
        pass

    def _post_detach_children(self, children):
        """Method call after detaching `children`."""
        pass

    def _pre_attach_children(self, children):
        """Method call before attaching `children`."""
        pass

    def _post_attach_children(self, children):
        """Method call after attaching `children`."""
        pass

    @property
    def path(self):
        """
        Path of this `Node`.

        >>> from anytree import Node
        >>> udo = Node("Udo")
        >>> marc = Node("Marc", parent=udo)
        >>> lian = Node("Lian", parent=marc)
        >>> udo.path
        (Node('/Udo'),)
        >>> marc.path
        (Node('/Udo'), Node('/Udo/Marc'))
        >>> lian.path
        (Node('/Udo'), Node('/Udo/Marc'), Node('/Udo/Marc/Lian'))
        """
        return self._path

    @property
    def _path(self):
        path = []
        node = self
        while node:
            path.insert(0, node)
            node = node.parent
        return tuple(path)

    @property
    def ancestors(self):
        """
        All parent nodes and their parent nodes.

        >>> from anytree import Node
        >>> udo = Node("Udo")
        >>> marc = Node("Marc", parent=udo)
        >>> lian = Node("Lian", parent=marc)
        >>> udo.ancestors
        ()
        >>> marc.ancestors
        (Node('/Udo'),)
        >>> lian.ancestors
        (Node('/Udo'), Node('/Udo/Marc'))
        """
        return self._path[:-1]

    @property
    def anchestors(self):
        """
        All parent nodes and their parent nodes - see :any:`ancestors`.

        The attribute `anchestors` is just a typo of `ancestors`. Please use `ancestors`.
        This attribute will be removed in the 2.0.0 release.
        """
        warnings.warn(".anchestors was a typo and will be removed in version 3.0.0", DeprecationWarning)
        return self.ancestors

    @property
    def descendants(self):
        """
        All child nodes and all their child nodes.

        >>> from anytree import Node
        >>> udo = Node("Udo")
        >>> marc = Node("Marc", parent=udo)
        >>> lian = Node("Lian", parent=marc)
        >>> loui = Node("Loui", parent=marc)
        >>> soe = Node("Soe", parent=lian)
        >>> udo.descendants
        (Node('/Udo/Marc'), Node('/Udo/Marc/Lian'), Node('/Udo/Marc/Lian/Soe'), Node('/Udo/Marc/Loui'))
        >>> marc.descendants
        (Node('/Udo/Marc/Lian'), Node('/Udo/Marc/Lian/Soe'), Node('/Udo/Marc/Loui'))
        >>> lian.descendants
        (Node('/Udo/Marc/Lian/Soe'),)
        """
        return tuple(PreOrderIter(self))[1:]

    @property
    def root(self):
        """
        Tree Root Node.

        >>> from anytree import Node
        >>> udo = Node("Udo")
        >>> marc = Node("Marc", parent=udo)
        >>> lian = Node("Lian", parent=marc)
        >>> udo.root
        Node('/Udo')
        >>> marc.root
        Node('/Udo')
        >>> lian.root
        Node('/Udo')
        """
        if self.parent:
            return self._path[0]
        else:
            return self

    @property
    def siblings(self):
        """
        Tuple of nodes with the same parent.

        >>> from anytree import Node
        >>> udo = Node("Udo")
        >>> marc = Node("Marc", parent=udo)
        >>> lian = Node("Lian", parent=marc)
        >>> loui = Node("Loui", parent=marc)
        >>> lazy = Node("Lazy", parent=marc)
        >>> udo.siblings
        ()
        >>> marc.siblings
        ()
        >>> lian.siblings
        (Node('/Udo/Marc/Loui'), Node('/Udo/Marc/Lazy'))
        >>> loui.siblings
        (Node('/Udo/Marc/Lian'), Node('/Udo/Marc/Lazy'))
        """
        parent = self.parent
        if parent is None:
            return tuple()
        else:
            return tuple([node for node in parent.children if node != self])

    @property
    def is_leaf(self):
        """
        `Node` has no children (External Node).

        >>> from anytree import Node
        >>> udo = Node("Udo")
        >>> marc = Node("Marc", parent=udo)
        >>> lian = Node("Lian", parent=marc)
        >>> udo.is_leaf
        False
        >>> marc.is_leaf
        False
        >>> lian.is_leaf
        True
        """
        return len(self.__children_) == 0

    @property
    def is_root(self):
        """
        `Node` is tree root.

        >>> from anytree import Node
        >>> udo = Node("Udo")
        >>> marc = Node("Marc", parent=udo)
        >>> lian = Node("Lian", parent=marc)
        >>> udo.is_root
        True
        >>> marc.is_root
        False
        >>> lian.is_root
        False
        """
        return self.parent is None

    @property
    def height(self):
        """
        Number of edges on the longest path to a leaf `Node`.

        >>> from anytree import Node
        >>> udo = Node("Udo")
        >>> marc = Node("Marc", parent=udo)
        >>> lian = Node("Lian", parent=marc)
        >>> udo.height
        2
        >>> marc.height
        1
        >>> lian.height
        0
        """
        if self.__children_:
            return max([child.height for child in self.__children_]) + 1
        else:
            return 0

    @property
    def depth(self):
        """
        Number of edges to the root `Node`.

        >>> from anytree import Node
        >>> udo = Node("Udo")
        >>> marc = Node("Marc", parent=udo)
        >>> lian = Node("Lian", parent=marc)
        >>> udo.depth
        0
        >>> marc.depth
        1
        >>> lian.depth
        2
        """
        return len(self._path) - 1

    def _pre_detach(self, parent):
        """Method call before detaching from `parent`."""
        pass

    def _post_detach(self, parent):
        """Method call after detaching from `parent`."""
        pass

    def _pre_attach(self, parent):
        """Method call before attaching to `parent`."""
        pass

    def _post_attach(self, parent):
        """Method call after attaching to `parent`."""
        pass



# -*- coding: utf-8 -*-

from .nodemixin import NodeMixin
from .util import _repr


class AnyNode(NodeMixin, object):

    def __init__(self, parent=None, **kwargs):
        u"""
        A generic tree node with any `kwargs`.

        >>> from anytree import AnyNode, RenderTree
        >>> root = AnyNode(id="root")
        >>> s0 = AnyNode(id="sub0", parent=root)
        >>> s0b = AnyNode(id="sub0B", parent=s0, foo=4, bar=109)
        >>> s0a = AnyNode(id="sub0A", parent=s0)
        >>> s1 = AnyNode(id="sub1", parent=root)
        >>> s1a = AnyNode(id="sub1A", parent=s1)
        >>> s1b = AnyNode(id="sub1B", parent=s1, bar=8)
        >>> s1c = AnyNode(id="sub1C", parent=s1)
        >>> s1ca = AnyNode(id="sub1Ca", parent=s1c)

        >>> root
        AnyNode(id='root')
        >>> s0
        AnyNode(id='sub0')
        >>> print(RenderTree(root))
        AnyNode(id='root')
        ├── AnyNode(id='sub0')
        │   ├── AnyNode(bar=109, foo=4, id='sub0B')
        │   └── AnyNode(id='sub0A')
        └── AnyNode(id='sub1')
            ├── AnyNode(id='sub1A')
            ├── AnyNode(bar=8, id='sub1B')
            └── AnyNode(id='sub1C')
                └── AnyNode(id='sub1Ca')
        """
        self.__dict__.update(kwargs)
        self.parent = parent

    def __repr__(self):
        return _repr(self)



import pathlib
from setuptools import setup, find_packages

# The directory containing this file
HERE = pathlib.Path(__file__).parent

# The text of the README file
README = (HERE / "README.md").read_text()

REQUIRES_PYTHON = '>=3.7.0'

setup(
    name="AGouTI",
    version="1.0.3",
    description="Annotation of Genomic and Transcriptomic Intervals using GTF/GFF files.",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/zywicki-lab/agouti",
    author="Jan Kosinski",
    python_requires=REQUIRES_PYTHON,
    author_email="jankos@amu.edu.pl",
    license="GNU General Public License",
    classifiers=[
        "License :: OSI Approved :: GNU General Public License (GPL)",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
    ],
    packages=find_packages(),
    include_package_data=True,
    package_data={'': ["sample_data.tar.gz"]},
    setup_requires=["numpy", "pandas"],
    install_requires=["numpy", "pandas"],
    scripts=['agouti_pkg/agouti'],

)



from agouti.gffutils import example_filename
import agouti.gffutils
import agouti.gffutils.biopython_integration as bp

def test_roundtrip():
    """
    Feature -> SeqFeature -> Feature should be invariant.
    """
    db_fname = gffutils.example_filename("gff_example1.gff3")
    db = gffutils.create_db(db_fname, ':memory:')
    feature = db['ENSMUSG00000033845']
    feature.keep_order = True
    dialect = feature.dialect
    s = bp.to_seqfeature(feature)
    assert s.location.start.position == feature.start - 1
    assert s.location.end.position == feature.stop
    f = bp.from_seqfeature(s, dialect=dialect, keep_order=True)
    assert feature == f



# -*- coding: utf-8 -*-
"""
Tree Rendering.

* :any:`RenderTree` using the following styles:
    * :any:`AsciiStyle`
    * :any:`ContStyle`
    * :any:`ContRoundStyle`
    * :any:`DoubleStyle`
"""

import collections

import agouti_pkg.six


Row = collections.namedtuple("Row", ("pre", "fill", "node"))


class AbstractStyle(object):

    def __init__(self, vertical, cont, end):
        """
        Tree Render Style.

        Args:

            vertical: Sign for vertical line.

            cont: Chars for a continued branch.

            end: Chars for the last branch.
        """
        super(AbstractStyle, self).__init__()
        self.vertical = vertical
        self.cont = cont
        self.end = end
        assert (len(cont) == len(vertical) and len(cont) == len(end)), (
            "'%s', '%s' and '%s' need to have equal length" % (vertical, cont,
                                                               end))

    @property
    def empty(self):
        """Empty string as placeholder."""
        return ' ' * len(self.end)

    def __repr__(self):
        classname = self.__class__.__name__
        return "%s()" % classname


class AsciiStyle(AbstractStyle):

    def __init__(self):
        """
        Ascii style.

        >>> from anytree import Node, RenderTree
        >>> root = Node("root")
        >>> s0 = Node("sub0", parent=root)
        >>> s0b = Node("sub0B", parent=s0)
        >>> s0a = Node("sub0A", parent=s0)
        >>> s1 = Node("sub1", parent=root)

        >>> print(RenderTree(root, style=AsciiStyle()))
        Node('/root')
        |-- Node('/root/sub0')
        |   |-- Node('/root/sub0/sub0B')
        |   +-- Node('/root/sub0/sub0A')
        +-- Node('/root/sub1')
        """
        super(AsciiStyle, self).__init__(u'|   ', u'|-- ', u'+-- ')


class ContStyle(AbstractStyle):

    def __init__(self):
        u"""
        Continued style, without gaps.

        >>> from anytree import Node, RenderTree
        >>> root = Node("root")
        >>> s0 = Node("sub0", parent=root)
        >>> s0b = Node("sub0B", parent=s0)
        >>> s0a = Node("sub0A", parent=s0)
        >>> s1 = Node("sub1", parent=root)

        >>> print(RenderTree(root, style=ContStyle()))
        Node('/root')
        ├── Node('/root/sub0')
        │   ├── Node('/root/sub0/sub0B')
        │   └── Node('/root/sub0/sub0A')
        └── Node('/root/sub1')
        """
        super(ContStyle, self).__init__(u'\u2502   ',
                                        u'\u251c\u2500\u2500 ',
                                        u'\u2514\u2500\u2500 ')


class ContRoundStyle(AbstractStyle):

    def __init__(self):
        u"""
        Continued style, without gaps, round edges.

        >>> from anytree import Node, RenderTree
        >>> root = Node("root")
        >>> s0 = Node("sub0", parent=root)
        >>> s0b = Node("sub0B", parent=s0)
        >>> s0a = Node("sub0A", parent=s0)
        >>> s1 = Node("sub1", parent=root)

        >>> print(RenderTree(root, style=ContRoundStyle()))
        Node('/root')
        ├── Node('/root/sub0')
        │   ├── Node('/root/sub0/sub0B')
        │   ╰── Node('/root/sub0/sub0A')
        ╰── Node('/root/sub1')
        """
        super(ContRoundStyle, self).__init__(u'\u2502   ',
                                             u'\u251c\u2500\u2500 ',
                                             u'\u2570\u2500\u2500 ')


class DoubleStyle(AbstractStyle):

    def __init__(self):
        u"""
        Double line style, without gaps.

        >>> from anytree import Node, RenderTree
        >>> root = Node("root")
        >>> s0 = Node("sub0", parent=root)
        >>> s0b = Node("sub0B", parent=s0)
        >>> s0a = Node("sub0A", parent=s0)
        >>> s1 = Node("sub1", parent=root)

        >>> print(RenderTree(root, style=DoubleStyle))
        Node('/root')
        ╠══ Node('/root/sub0')
        ║   ╠══ Node('/root/sub0/sub0B')
        ║   ╚══ Node('/root/sub0/sub0A')
        ╚══ Node('/root/sub1')

        """
        super(DoubleStyle, self).__init__(u'\u2551   ',
                                          u'\u2560\u2550\u2550 ',
                                          u'\u255a\u2550\u2550 ')


@agouti_pkg.six.python_2_unicode_compatible
class RenderTree(object):

    def __init__(self, node, style=ContStyle(), childiter=list):
        u"""
        Render tree starting at `node`.

        Keyword Args:
            style (AbstractStyle): Render Style.
            childiter: Child iterator.

        :any:`RenderTree` is an iterator, returning a tuple with 3 items:

        `pre`
            tree prefix.

        `fill`
            filling for multiline entries.

        `node`
            :any:`NodeMixin` object.

        It is up to the user to assemble these parts to a whole.

        >>> from anytree import Node, RenderTree
        >>> root = Node("root", lines=["c0fe", "c0de"])
        >>> s0 = Node("sub0", parent=root, lines=["ha", "ba"])
        >>> s0b = Node("sub0B", parent=s0, lines=["1", "2", "3"])
        >>> s0a = Node("sub0A", parent=s0, lines=["a", "b"])
        >>> s1 = Node("sub1", parent=root, lines=["Z"])

        Simple one line:

        >>> for pre, _, node in RenderTree(root):
        ...     print("%s%s" % (pre, node.name))
        root
        ├── sub0
        │   ├── sub0B
        │   └── sub0A
        └── sub1

        Multiline:

        >>> for pre, fill, node in RenderTree(root):
        ...     print("%s%s" % (pre, node.lines[0]))
        ...     for line in node.lines[1:]:
        ...         print("%s%s" % (fill, line))
        c0fe
        c0de
        ├── ha
        │   ba
        │   ├── 1
        │   │   2
        │   │   3
        │   └── a
        │       b
        └── Z

        The `childiter` is responsible for iterating over child nodes at the
        same level. An reversed order can be achived by using `reversed`.

        >>> for row in RenderTree(root, childiter=reversed):
        ...     print("%s%s" % (row.pre, row.node.name))
        root
        ├── sub1
        └── sub0
            ├── sub0A
            └── sub0B

        Or writing your own sort function:

        >>> def mysort(items):
        ...     return sorted(items, key=lambda item: item.name)
        >>> for row in RenderTree(root, childiter=mysort):
        ...     print("%s%s" % (row.pre, row.node.name))
        root
        ├── sub0
        │   ├── sub0A
        │   └── sub0B
        └── sub1

        :any:`by_attr` simplifies attribute rendering and supports multiline:

        >>> print(RenderTree(root).by_attr())
        root
        ├── sub0
        │   ├── sub0B
        │   └── sub0A
        └── sub1
        >>> print(RenderTree(root).by_attr("lines"))
        c0fe
        c0de
        ├── ha
        │   ba
        │   ├── 1
        │   │   2
        │   │   3
        │   └── a
        │       b
        └── Z
        """
        if not isinstance(style, AbstractStyle):
            style = style()
        self.node = node
        self.style = style
        self.childiter = childiter

    def __iter__(self):
        return self.__next(self.node, tuple())

    def __next(self, node, continues):
        yield RenderTree.__item(node, continues, self.style)
        children = node.children
        if children:
            lastidx = len(children) - 1
            for idx, child in enumerate(self.childiter(children)):
                for grandchild in self.__next(child, continues + (idx != lastidx, )):
                    yield grandchild

    @staticmethod
    def __item(node, continues, style):
        if not continues:
            return Row(u'', u'', node)
        else:
            items = [style.vertical if cont else style.empty for cont in continues]
            indent = ''.join(items[:-1])
            branch = style.cont if continues[-1] else style.end
            pre = indent + branch
            fill = ''.join(items)
            return Row(pre, fill, node)

    def __str__(self):
        lines = ["%s%r" % (pre, node) for pre, _, node in self]
        return "\n".join(lines)

    def __repr__(self):
        classname = self.__class__.__name__
        args = [repr(self.node),
                "style=%s" % repr(self.style),
                "childiter=%s" % repr(self.childiter)]
        return "%s(%s)" % (classname, ", ".join(args))

    def by_attr(self, attrname="name"):
        """Return rendered tree with node attribute `attrname`."""
        def get():
            for pre, fill, node in self:
                attr = getattr(node, attrname, "")
                if isinstance(attr, (list, tuple)):
                    lines = attr
                else:
                    lines = str(attr).split("\n")
                yield u"%s%s" % (pre, lines[0])
                for line in lines[1:]:
                    yield u"%s%s" % (fill, line)
        return "\n".join(get())
